from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, List, Literal, Type, Union, overload
from urllib import parse

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.pydantic import AnyUrl
from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.datasource.fluent.interfaces import (
    DataAsset,
    TestConnectionError,
)
from great_expectations.datasource.fluent.sql_datasource import (
    QueryAsset as SqlQueryAsset,
)
from great_expectations.datasource.fluent.sql_datasource import (
    SQLDatasource,
)
from great_expectations.datasource.fluent.sql_datasource import (
    TableAsset as SqlTableAsset,
)

if TYPE_CHECKING:
    from sqlalchemy.sql import quoted_name  # noqa: TID251 # type-checking only

    from great_expectations.compatibility import sqlalchemy
    from great_expectations.compatibility.pydantic.networks import Parts
    from great_expectations.core.config_provider import _ConfigurationProvider


def _parse_param_from_query_string(param: str, query: str) -> str | None:
    url_components = parse.urlparse(query)
    path = str(url_components.path)
    parse_results: dict[str, list[str]] = parse.parse_qs(path)
    path_results = parse_results.get(param, [])

    if not path_results:
        return None
    if len(path_results) > 1:
        raise ValueError(f"Only one `{param}` query entry is allowed")  # noqa: TRY003 # FIXME CoP
    return path_results[0]


class _UrlQueryError(pydantic.UrlError):
    """
    Custom Pydantic error for missing query in DatabricksDsn.
    """

    code = "url.query"
    msg_template = "URL query is invalid or missing"


class _UrlHttpPathError(pydantic.UrlError):
    """
    Custom Pydantic error for missing http_path in DatabricksDsn query.
    """

    code = "url.query.http_path"
    msg_template = "'http_path' query param is invalid or missing"


class _UrlCatalogError(pydantic.UrlError):
    """
    Custom Pydantic error for missing catalog in DatabricksDsn query.
    """

    code = "url.query.catalog"
    msg_template = "'catalog' query param is invalid or missing"


class _UrlSchemaError(pydantic.UrlError):
    """
    Custom Pydantic error for missing schema in DatabricksDsn query.
    """

    code = "url.query.schema"
    msg_template = "'schema' query param is invalid or missing"


class DatabricksDsn(AnyUrl):
    allowed_schemes = {
        "databricks",
    }
    query: str  # if query is not provided, validate_parts() will raise an error

    @classmethod
    @override
    def validate_parts(cls, parts: Parts, validate_port: bool = True) -> Parts:
        """
        Overridden to validate additional fields outside of scheme (which is performed by AnyUrl).
        """
        query = parts["query"]
        if query is None:
            raise _UrlQueryError()

        http_path = _parse_param_from_query_string(param="http_path", query=query)
        if http_path is None:
            raise _UrlHttpPathError()

        catalog = _parse_param_from_query_string(param="catalog", query=query)
        if catalog is None:
            raise _UrlCatalogError()

        schema = _parse_param_from_query_string(param="schema", query=query)
        if schema is None:
            raise _UrlSchemaError()

        return AnyUrl.validate_parts(parts=parts, validate_port=validate_port)

    @overload
    @classmethod
    def parse_url(
        cls, url: ConfigStr, config_provider: _ConfigurationProvider = ...
    ) -> DatabricksDsn: ...

    @overload
    @classmethod
    def parse_url(
        cls, url: str, config_provider: _ConfigurationProvider | None = ...
    ) -> DatabricksDsn: ...

    @classmethod
    def parse_url(
        cls, url: ConfigStr | str, config_provider: _ConfigurationProvider | None = None
    ) -> DatabricksDsn:
        if isinstance(url, ConfigStr):
            assert config_provider, "`config_provider` must be provided"
            url = url.get_config_value(config_provider=config_provider)
        parsed_url = pydantic.parse_obj_as(DatabricksDsn, url)
        return parsed_url


class DatabricksTableAsset(SqlTableAsset):
    @pydantic.validator("table_name")
    @override
    def _resolve_quoted_name(cls, table_name: str) -> str | quoted_name:
        table_name_is_quoted: bool = cls._is_bracketed_by_quotes(table_name)

        from great_expectations.compatibility import sqlalchemy

        if sqlalchemy.quoted_name:  # type: ignore[truthy-function] # FIXME CoP
            if isinstance(table_name, sqlalchemy.quoted_name):
                return table_name

            if table_name_is_quoted:
                # https://docs.sqlalchemy.org/en/20/core/sqlelement.html#sqlalchemy.sql.expression.quoted_name.quote
                # Remove the quotes and add them back using the sqlalchemy.quoted_name function
                # TODO: We need to handle nested quotes
                table_name = table_name.strip("`")

            return sqlalchemy.quoted_name(
                value=table_name,
                quote=table_name_is_quoted,
            )
        return table_name

    @staticmethod
    @override
    def _is_bracketed_by_quotes(target: str) -> bool:
        """Returns True if the target string is bracketed by quotes.

        Arguments:
            target: A string to check if it is bracketed by quotes.

        Returns:
            True if the target string is bracketed by quotes.
        """
        # TODO: what todo with regular quotes? Error? Warn? "Fix"?
        return target.startswith("`") and target.endswith("`")


@public_api
class DatabricksSQLDatasource(SQLDatasource):
    """Adds a DatabricksSQLDatasource to the data context.

    Args:
        name: The name of this DatabricksSQL datasource.
        connection_string: The SQLAlchemy connection string used to connect to the Databricks SQL database.
            For example: "databricks://token:<token>@<host>:<port>?http_path=<http_path>&catalog=<catalog>&schema=<schema>""
        assets: An optional dictionary whose keys are TableAsset or QueryAsset names and whose values
            are TableAsset or QueryAsset objects.
    """  # noqa: E501 # FIXME CoP

    # class var definitions
    asset_types: ClassVar[List[Type[DataAsset]]] = [DatabricksTableAsset, SqlQueryAsset]

    type: Literal["databricks_sql"] = "databricks_sql"  # type: ignore[assignment] # FIXME CoP
    connection_string: Union[ConfigStr, DatabricksDsn]

    # These are instance var because ClassVars can't contain Type variables. See
    # https://peps.python.org/pep-0526/#class-and-instance-variable-annotations
    _TableAsset: Type[SqlTableAsset] = pydantic.PrivateAttr(DatabricksTableAsset)

    @override
    def test_connection(self, test_assets: bool = True) -> None:
        try:
            super().test_connection(test_assets)
        except TestConnectionError as e:
            nested_exception = None
            if e.__cause__ and e.__cause__.__cause__:
                nested_exception = e.__cause__.__cause__

            # Raise specific error informing how to install dependencies only if relevant
            if isinstance(nested_exception, sa.exc.NoSuchModuleError):
                raise TestConnectionError(  # noqa: TRY003 # FIXME CoP
                    "Could not connect to Databricks - please ensure you've installed necessary dependencies with `pip install great_expectations[databricks]`."  # noqa: E501 # FIXME CoP
                ) from e
            raise e  # noqa: TRY201 # FIXME CoP

    @override
    def _create_engine(self) -> sqlalchemy.Engine:
        model_dict = self.dict(
            exclude=self._get_exec_engine_excludes(),
            config_provider=self._config_provider,
        )

        connection_string = model_dict.pop("connection_string")
        # is connection_string was a ConfigStr it's parts will not have been validated yet
        if not isinstance(connection_string, DatabricksDsn):
            connection_string = DatabricksDsn.parse_url(
                url=connection_string, config_provider=self._config_provider
            )

        kwargs = model_dict.pop("kwargs", {})

        http_path = _parse_param_from_query_string(param="http_path", query=connection_string.query)
        assert http_path, "Presence of http_path query string is guaranteed due to prior validation"

        # Databricks connection is a bit finicky - the http_path portion of the connection string needs to be passed in connect_args  # noqa: E501 # FIXME CoP
        connect_args = {"http_path": http_path}
        return sa.create_engine(connection_string, connect_args=connect_args, **kwargs)



from typing import Literal, Optional

from great_expectations.compatibility.pydantic import StrictStr
from great_expectations.expectations.model_field_types import ConditionParser
from great_expectations.metrics.metric import Metric, NonEmptyString, _MetricResult


class ColumnPairMetric(Metric[_MetricResult], kw_only=True):
    column_A: NonEmptyString
    column_B: NonEmptyString
    ignore_row_if: Literal["both_values_are_missing", "either_value_is_missing", "neither"] = (
        "both_values_are_missing"
    )
    row_condition: Optional[StrictStr] = None
    condition_parser: Optional[ConditionParser] = None



import warnings

import boto3
from awsglue.context import GlueContext
from pyspark.context import SparkContext

from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context import get_context
from great_expectations.data_context.types.base import (
    DataContextConfig,
    S3StoreBackendDefaults,
)

yaml = YAMLHandler()

# needed because GlueContext(sc) function emits the following FutureWarning: Deprecated in 3.0.0. Use SparkSession.builder.getOrCreate() instead.
with warnings.catch_warnings():
    warnings.simplefilter(action="ignore", category=FutureWarning)
    sc = SparkContext.getOrCreate()
    glueContext = GlueContext(sc)
    spark = glueContext.spark_session

s3_client = boto3.client("s3")
response = s3_client.get_object(
    Bucket="bucket", Key="bucket/great_expectations/great_expectations.yml"
)
config_file = yaml.load(response["Body"])

config = DataContextConfig(
    config_version=config_file["config_version"],
    datasources=config_file["datasources"],
    expectations_store_name=config_file["expectations_store_name"],
    validation_results_store_name=config_file["validation_results_store_name"],
    plugins_directory="/great_expectations/plugins",
    stores=config_file["stores"],
    data_docs_sites=config_file["data_docs_sites"],
    config_variables_file_path=config_file["config_variables_file_path"],
    anonymous_usage_statistics=config_file["anonymous_usage_statistics"],
    checkpoint_store_name=config_file["checkpoint_store_name"],
    store_backend_defaults=S3StoreBackendDefaults(
        default_bucket_name=config_file["data_docs_sites"]["s3_site"]["store_backend"]["bucket"]
    ),
)
context_gx = get_context(project_config=config)

expectation_suite_name = "suite_name"
suite = context_gx.suites.add(ExpectationSuite(name=expectation_suite_name))
batch_request = RuntimeBatchRequest(
    datasource_name="spark_s3",
    data_asset_name="datafile_name",
    batch_identifiers={"runtime_batch_identifier_name": "default_identifier"},
    data_connector_name="default_inferred_data_connector_name",
    runtime_parameters={"path": "s3a://bucket_name/path_to_file.format"},
)
validator = context_gx.get_validator(
    batch_request=batch_request,
    expectation_suite_name=expectation_suite_name,
)
print(validator.head())
validator.expect_column_values_to_not_be_null(column="passenger_count")  ## add some test
validator.save_expectation_suite(discard_failed_expectations=False)

checkpoint_config = {
    "class_name": "Checkpoint",
    "validations": [
        {
            "batch_request": batch_request,
            "expectation_suite_name": expectation_suite_name,
        }
    ],
}

checkpoint = context_gx.add_or_update_checkpoint(
    f"_tmp_checkpoint_{expectation_suite_name}", **checkpoint_config
)
results = checkpoint.run(result_format="SUMMARY", run_name="test")
validation_result_identifier = results.list_validation_result_identifiers()[0]



from .batch.batch_column_types import BatchColumnTypes
from .batch.row_count import BatchRowCount
from .batch.sample_values import SampleValues
from .column.aggregate_non_null_count import ColumnAggregateNonNullCount
from .column.descriptive_stats import ColumnDescriptiveStats
from .column.distinct_values import ColumnDistinctValues
from .column.distinct_values_count import ColumnDistinctValuesCount
from .column.mean import ColumnMean
from .column.null_count import ColumnNullCount
from .column.sample_values import ColumnSampleValues
from .column.values_match_regex_count import ColumnValuesMatchRegexCount
from .column.values_match_regex_values import ColumnValuesMatchRegexValues
from .column.values_non_null import ColumnValuesNonNull, ColumnValuesNonNullCount
from .column.values_not_match_regex_count import ColumnValuesNotMatchRegexCount
from .column.values_not_match_regex_values import ColumnValuesNotMatchRegexValues
from .column_pair.values_in_set import ColumnPairValuesInSetUnexpectedCount
from .metric import Metric
from .multi_column.sum_equal import MultiColumnSumEqualUnexpectedCount
from .query.data_source_table import QueryDataSourceTable
from .query.row_count import QueryRowCount



from __future__ import annotations

import copy
import json
import logging
import pathlib
import sys
import time
import traceback
from collections import defaultdict
from typing import TYPE_CHECKING, Final, List, Optional, Union

from great_expectations.core.expectation_diagnostics.expectation_diagnostics import (
    ExpectationDiagnostics,
)
from great_expectations.core.expectation_diagnostics.expectation_test_data_cases import (
    ExpectationLegacyTestCaseAdapter,
    ExpectationTestDataCases,
    TestBackend,
)
from great_expectations.core.expectation_diagnostics.supporting_types import (
    AugmentedLibraryMetadata,
    ExpectationBackendTestResultCounts,
    ExpectationDescriptionDiagnostics,
    ExpectationDiagnosticMaturityMessages,
    ExpectationErrorDiagnostics,
    ExpectationExecutionEngineDiagnostics,
    ExpectationMetricDiagnostics,
    ExpectationRendererDiagnostics,
    ExpectationTestDiagnostics,
    Maturity,
    RendererTestDiagnostics,
)
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.expectations.registry import (
    _registered_metrics,
    _registered_renderers,
)
from great_expectations.render import (
    CollapseContent,
    LegacyDiagnosticRendererType,
    LegacyRendererType,
    RenderedAtomicContent,
    RenderedContentBlockContainer,
    RenderedGraphContent,
    RenderedStringTemplateContent,
    RenderedTableContent,
    ValueListContent,
)
from great_expectations.self_check.util import (
    evaluate_json_test_v3_api,
    generate_dataset_name_from_expectation_name,
    generate_expectation_tests,
)
from great_expectations.util import camel_to_snake

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )
    from great_expectations.expectations.expectation import Expectation
    from great_expectations.validator.validator import ValidationDependencies

_TEST_DEFS_DIR: Final = pathlib.Path(
    __file__, "..", "..", "..", "..", "tests", "test_definitions"
).resolve()


class ExpectationDoctor:
    def __init__(self, expectation: Expectation) -> None:
        self._expectation = expectation

    def print_diagnostic_checklist(
        self,
        diagnostics: Optional[ExpectationDiagnostics] = None,
        show_failed_tests: bool = False,
        backends: Optional[List[str]] = None,
        show_debug_messages: bool = False,
    ) -> str:
        if diagnostics is None:
            debug_logger = None
            if show_debug_messages:
                debug_logger = logging.getLogger()
                chandler = logging.StreamHandler(stream=sys.stdout)
                chandler.setLevel(logging.DEBUG)
                chandler.setFormatter(
                    logging.Formatter(
                        "%(asctime)s - %(levelname)s - %(message)s", "%Y-%m-%dT%H:%M:%S"
                    )
                )
                debug_logger.addHandler(chandler)
                debug_logger.setLevel(logging.DEBUG)

            diagnostics = self.run_diagnostics(
                debug_logger=debug_logger, only_consider_these_backends=backends
            )
        if show_failed_tests:
            for test in diagnostics.tests:
                if test.test_passed is False:
                    print(f"=== {test.test_title} ({test.backend}) ===\n")
                    print(f"{80 * '='}\n")

        checklist: str = diagnostics.generate_checklist()
        print(checklist)

        return checklist

    def run_diagnostics(  # noqa: PLR0913 # FIXME CoP
        self,
        raise_exceptions_for_backends: bool = False,
        ignore_suppress: bool = False,
        ignore_only_for: bool = False,
        for_gallery: bool = False,
        debug_logger: Optional[logging.Logger] = None,
        only_consider_these_backends: Optional[List[str]] = None,
        context: Optional[AbstractDataContext] = None,
    ) -> ExpectationDiagnostics:
        if debug_logger is not None:
            _debug = lambda x: debug_logger.debug(  # noqa: E731 # FIXME CoP
                f"(run_diagnostics) {x}"
            )
            _error = lambda x: debug_logger.error(  # noqa: E731 # FIXME CoP
                f"(run_diagnostics) {x}"
            )
        else:
            _debug = lambda x: x  # noqa: E731 # FIXME CoP
            _error = lambda x: x  # noqa: E731 # FIXME CoP

        library_metadata: AugmentedLibraryMetadata = self._get_augmented_library_metadata()
        examples: List[ExpectationTestDataCases] = self._get_examples(
            return_only_gallery_examples=False
        )
        gallery_examples: List[ExpectationTestDataCases] = []
        for example in examples:
            _tests_to_include = [test for test in example.tests if test.include_in_gallery]
            example = copy.deepcopy(example)  # noqa: PLW2901 # FIXME CoP
            if _tests_to_include:
                example.tests = _tests_to_include
                gallery_examples.append(example)

        description_diagnostics: ExpectationDescriptionDiagnostics = (
            self._get_description_diagnostics()
        )

        _expectation_config: Optional[ExpectationConfiguration] = (
            self._get_expectation_configuration_from_examples(examples)
        )
        if not _expectation_config:
            _error(
                f"Was NOT able to get Expectation configuration for {self._expectation.expectation_type}. "  # noqa: E501 # FIXME CoP
                "Is there at least one sample test where 'success' is True?"
            )
        metric_diagnostics_list: List[ExpectationMetricDiagnostics] = (
            self._get_metric_diagnostics_list(
                expectation_config=_expectation_config,
            )
        )

        introspected_execution_engines: ExpectationExecutionEngineDiagnostics = (
            self._get_execution_engine_diagnostics(
                metric_diagnostics_list=metric_diagnostics_list,
                registered_metrics=_registered_metrics,
            )
        )
        engines_implemented = [
            e.replace("ExecutionEngine", "")
            for e, i in introspected_execution_engines.items()
            if i is True
        ]
        _debug(
            f"Implemented engines for {self._expectation.expectation_type}: {', '.join(engines_implemented)}"  # noqa: E501 # FIXME CoP
        )

        _debug("Getting test results")
        test_results: List[ExpectationTestDiagnostics] = self._get_test_results(
            expectation_type=description_diagnostics.snake_name,
            test_data_cases=examples,
            execution_engine_diagnostics=introspected_execution_engines,
            raise_exceptions_for_backends=raise_exceptions_for_backends,
            ignore_suppress=ignore_suppress,
            ignore_only_for=ignore_only_for,
            debug_logger=debug_logger,
            only_consider_these_backends=only_consider_these_backends,
            context=context,
        )

        backend_test_result_counts: List[ExpectationBackendTestResultCounts] = (
            ExpectationDiagnostics._get_backends_from_test_results(test_results)
        )

        renderers: List[ExpectationRendererDiagnostics] = self._get_renderer_diagnostics(
            expectation_type=description_diagnostics.snake_name,
            test_diagnostics=test_results,
            registered_renderers=_registered_renderers,  # type: ignore[arg-type] # FIXME CoP
        )

        maturity_checklist: ExpectationDiagnosticMaturityMessages = self._get_maturity_checklist(
            library_metadata=library_metadata,
            description=description_diagnostics,
            examples=examples,
            tests=test_results,
            backend_test_result_counts=backend_test_result_counts,
        )

        coverage_score: float = self._get_coverage_score(
            backend_test_result_counts=backend_test_result_counts,
            execution_engines=introspected_execution_engines,
        )

        _debug(f"coverage_score: {coverage_score} for {self._expectation.expectation_type}")

        # Set final maturity level based on status of all checks
        library_metadata.maturity = self._get_final_maturity_level(
            maturity_checklist=maturity_checklist
        )

        # Set the errors found when running tests
        errors = [
            test_result.error_diagnostics
            for test_result in test_results
            if test_result.error_diagnostics
        ]

        # If run for the gallery, don't include a bunch of stuff
        #   - Don't set examples and test_results to empty lists here since these
        #     returned attributes will be needed to re-calculate the maturity
        #     checklist later (after merging results from different runs of the
        #     build_gallery.py script per backend)
        if for_gallery:
            gallery_examples = []
            renderers = []
            errors = []

        return ExpectationDiagnostics(
            library_metadata=library_metadata,
            examples=examples,
            gallery_examples=gallery_examples,
            description=description_diagnostics,
            renderers=renderers,
            metrics=metric_diagnostics_list,
            execution_engines=introspected_execution_engines,
            tests=test_results,
            backend_test_result_counts=backend_test_result_counts,
            maturity_checklist=maturity_checklist,
            errors=errors,
            coverage_score=coverage_score,
        )

    def _get_augmented_library_metadata(self):
        """Introspect the Expectation's library_metadata object (if it exists), and augment it with additional information."""  # noqa: E501 # FIXME CoP

        augmented_library_metadata = {
            "maturity": Maturity.CONCEPT_ONLY,
            "tags": [],
            "contributors": [],
            "requirements": [],
            "library_metadata_passed_checks": False,
            "has_full_test_suite": False,
            "manually_reviewed_code": False,
        }
        required_keys = {"contributors", "tags"}
        allowed_keys = {
            "contributors",
            "has_full_test_suite",
            "manually_reviewed_code",
            "maturity",
            "requirements",
            "tags",
        }
        problems = []

        if hasattr(self._expectation, "library_metadata"):
            augmented_library_metadata.update(self._expectation.library_metadata)
            keys = set(self._expectation.library_metadata.keys())
            missing_required_keys = required_keys - keys
            forbidden_keys = keys - allowed_keys

            if missing_required_keys:
                problems.append(f"Missing required key(s): {sorted(missing_required_keys)}")
            if forbidden_keys:
                problems.append(f"Extra key(s) found: {sorted(forbidden_keys)}")
            if not isinstance(augmented_library_metadata["requirements"], list):
                problems.append("library_metadata['requirements'] is not a list ")
            if not problems:
                augmented_library_metadata["library_metadata_passed_checks"] = True
        else:
            problems.append("No library_metadata attribute found")

        augmented_library_metadata["problems"] = problems
        return AugmentedLibraryMetadata.from_legacy_dict(augmented_library_metadata)

    def _get_maturity_checklist(
        self,
        library_metadata: Union[AugmentedLibraryMetadata, ExpectationDescriptionDiagnostics],
        description: ExpectationDescriptionDiagnostics,
        examples: List[ExpectationTestDataCases],
        tests: List[ExpectationTestDiagnostics],
        backend_test_result_counts: List[ExpectationBackendTestResultCounts],
    ) -> ExpectationDiagnosticMaturityMessages:
        """Generate maturity checklist messages"""
        experimental_checks = []
        beta_checks = []
        production_checks = []

        experimental_checks.append(ExpectationDiagnostics._check_library_metadata(library_metadata))
        experimental_checks.append(ExpectationDiagnostics._check_docstring(description))
        experimental_checks.append(ExpectationDiagnostics._check_example_cases(examples, tests))
        experimental_checks.append(
            ExpectationDiagnostics._check_core_logic_for_at_least_one_execution_engine(
                backend_test_result_counts
            )
        )
        beta_checks.append(
            ExpectationDiagnostics._check_input_validation(self._expectation, examples)
        )
        beta_checks.append(ExpectationDiagnostics._check_renderer_methods(self._expectation))
        beta_checks.append(
            ExpectationDiagnostics._check_core_logic_for_all_applicable_execution_engines(
                backend_test_result_counts
            )
        )

        production_checks.append(ExpectationDiagnostics._check_full_test_suite(library_metadata))
        production_checks.append(ExpectationDiagnostics._check_manual_code_review(library_metadata))

        return ExpectationDiagnosticMaturityMessages(
            experimental=experimental_checks,
            beta=beta_checks,
            production=production_checks,
        )

    @staticmethod
    def _get_coverage_score(
        backend_test_result_counts: List[ExpectationBackendTestResultCounts],
        execution_engines: ExpectationExecutionEngineDiagnostics,
    ) -> float:
        """Generate coverage score"""
        _total_passed = 0
        _total_failed = 0
        _num_backends = 0
        _num_engines = sum(x for x in execution_engines.values() if x)
        for result in backend_test_result_counts:
            _num_backends += 1
            _total_passed += result.num_passed
            _total_failed += result.num_failed

        coverage_score = _num_backends + _num_engines + _total_passed - (1.5 * _total_failed)

        return coverage_score

    @staticmethod
    def _get_final_maturity_level(
        maturity_checklist: ExpectationDiagnosticMaturityMessages,
    ) -> Maturity:
        """Get final maturity level based on status of all checks"""
        maturity = ""
        all_experimental = all(check.passed for check in maturity_checklist.experimental)
        all_beta = all(check.passed for check in maturity_checklist.beta)
        all_production = all(check.passed for check in maturity_checklist.production)
        if all_production and all_beta and all_experimental:
            maturity = Maturity.PRODUCTION
        elif all_beta and all_experimental:
            maturity = Maturity.BETA
        else:
            maturity = Maturity.EXPERIMENTAL

        return maturity

    def _get_examples_from_json(self):
        """Only meant to be called by self._get_examples"""
        results = []
        found = next(_TEST_DEFS_DIR.rglob(f"**/{self._expectation.expectation_type}.json"), None)
        if found:
            with open(found) as fp:
                data = json.load(fp)
            results = data["datasets"]
        return results

    def _get_examples(  # noqa: C901 #  too complex
        self, return_only_gallery_examples: bool = True
    ) -> List[ExpectationTestDataCases]:
        """
        Get a list of examples from the object's `examples` member variable.

        For core expectations, the examples are found in tests/test_definitions/

        :param return_only_gallery_examples: if True, include only test examples where `include_in_gallery` is true
        :return: list of examples or [], if no examples exist
        """  # noqa: E501 # FIXME CoP
        # Currently, only community contrib expectations have an examples attribute
        all_examples: List[dict] = self._expectation.examples or self._get_examples_from_json()

        included_examples = []
        for i, example in enumerate(all_examples, 1):
            included_test_cases = []
            # As of commit 7766bb5caa4e0 on 1/28/22, only_for does not need to be applied to individual tests  # noqa: E501 # FIXME CoP
            # See:
            #   - https://github.com/great-expectations/great_expectations/blob/7766bb5caa4e0e5b22fa3b3a5e1f2ac18922fdeb/tests/test_definitions/column_map_expectations/expect_column_values_to_be_unique.json#L174
            #   - https://github.com/great-expectations/great_expectations/pull/4073
            top_level_only_for = example.get("only_for")
            top_level_suppress_test_for = example.get("suppress_test_for")
            for test in example["tests"]:
                if (
                    test.get("include_in_gallery") == True  # noqa: E712 # FIXME CoP
                    or return_only_gallery_examples == False  # noqa: E712 # FIXME CoP
                ):
                    copied_test = copy.deepcopy(test)
                    if top_level_only_for:
                        if "only_for" not in copied_test:
                            copied_test["only_for"] = top_level_only_for
                        else:
                            copied_test["only_for"].extend(top_level_only_for)
                    if top_level_suppress_test_for:
                        if "suppress_test_for" not in copied_test:
                            copied_test["suppress_test_for"] = top_level_suppress_test_for
                        else:
                            copied_test["suppress_test_for"].extend(top_level_suppress_test_for)
                    included_test_cases.append(ExpectationLegacyTestCaseAdapter(**copied_test))

            # If at least one ExpectationTestCase from the ExpectationTestDataCases was selected,
            # then keep a copy of the ExpectationTestDataCases including data and the selected ExpectationTestCases.  # noqa: E501 # FIXME CoP
            if len(included_test_cases) > 0:
                copied_example = copy.deepcopy(example)
                copied_example["tests"] = included_test_cases
                copied_example.pop("_notes", None)
                copied_example.pop("only_for", None)
                copied_example.pop("suppress_test_for", None)
                if "test_backends" in copied_example:
                    copied_example["test_backends"] = [
                        TestBackend(**tb) for tb in copied_example["test_backends"]
                    ]

                if "dataset_name" not in copied_example:
                    dataset_name = generate_dataset_name_from_expectation_name(
                        dataset=copied_example,
                        expectation_type=self._expectation.expectation_type,
                        index=i,
                    )
                    copied_example["dataset_name"] = dataset_name

                included_examples.append(ExpectationTestDataCases(**copied_example))

        return included_examples

    def _get_docstring_and_short_description(self) -> tuple[str, str]:
        """Conveninence method to get the Exepctation's docstring and first line"""

        if self._expectation.__doc__ is not None:
            docstring = self._expectation.__doc__
            short_description = next(line for line in self._expectation.__doc__.split("\n") if line)
        else:
            docstring = ""
            short_description = ""

        return docstring, short_description

    def _get_description_diagnostics(self) -> ExpectationDescriptionDiagnostics:
        """Introspect the Expectation and create its ExpectationDescriptionDiagnostics object"""

        camel_name = self._expectation.__class__.__name__
        snake_name = camel_to_snake(camel_name)
        docstring, short_description = self._get_docstring_and_short_description()

        return ExpectationDescriptionDiagnostics(
            **{
                "camel_name": camel_name,
                "snake_name": snake_name,
                "short_description": short_description,
                "docstring": docstring,
            }
        )

    def _get_expectation_configuration_from_examples(  # noqa: C901 #  too complex
        self,
        examples: List[ExpectationTestDataCases],
    ) -> Optional[ExpectationConfiguration]:
        """Return an ExpectationConfiguration instance using test input expected to succeed"""
        if examples:
            for example in examples:
                tests = example.tests
                if tests:
                    for test in tests:
                        if test.output.get("success"):
                            return ExpectationConfiguration(
                                type=self._expectation.expectation_type,
                                kwargs=test.input,
                            )

            # There is no sample test where `success` is True, or there are no tests
            for example in examples:
                tests = example.tests
                if tests:
                    for test in tests:
                        if test.input:
                            return ExpectationConfiguration(
                                type=self._expectation.expectation_type,
                                kwargs=test.input,
                            )
        return None

    @staticmethod
    def _get_execution_engine_diagnostics(
        metric_diagnostics_list: List[ExpectationMetricDiagnostics],
        registered_metrics: dict,
        execution_engine_names: Optional[List[str]] = None,
    ) -> ExpectationExecutionEngineDiagnostics:
        """Check to see which execution_engines are fully supported for this Expectation.

        In order for a given execution engine to count, *every* metric must have support on that execution engines.
        """  # noqa: E501 # FIXME CoP
        if not execution_engine_names:
            execution_engine_names = [
                "PandasExecutionEngine",
                "SqlAlchemyExecutionEngine",
                "SparkDFExecutionEngine",
            ]

        execution_engines = {}
        for provider in execution_engine_names:
            all_true = True
            if not metric_diagnostics_list:
                all_true = False
            for metric_diagnostics in metric_diagnostics_list:
                try:
                    has_provider = (
                        provider in registered_metrics[metric_diagnostics.name]["providers"]
                    )
                    if not has_provider:
                        all_true = False
                        break
                except KeyError:
                    # https://github.com/great-expectations/great_expectations/blob/abd8f68a162eaf9c33839d2c412d8ba84f5d725b/great_expectations/expectations/core/expect_table_row_count_to_equal_other_table.py#L174-L181
                    # expect_table_row_count_to_equal_other_table does tricky things and replaces
                    # registered metric "table.row_count" with "table.row_count.self" and "table.row_count.other"  # noqa: E501 # FIXME CoP
                    if "table.row_count" in metric_diagnostics.name:
                        continue

            execution_engines[provider] = all_true

        return ExpectationExecutionEngineDiagnostics(**execution_engines)

    def _get_metric_diagnostics_list(
        self,
        expectation_config: Optional[ExpectationConfiguration],
    ) -> List[ExpectationMetricDiagnostics]:
        """Check to see which Metrics are upstream validation_dependencies for this Expectation."""

        # NOTE: Abe 20210102: Strictly speaking, identifying upstream metrics shouldn't need to rely on an expectation config.  # noqa: E501 # FIXME CoP
        # There's probably some part of get_validation_dependencies that can be factored out to remove the dependency.  # noqa: E501 # FIXME CoP

        if not expectation_config:
            return []

        validation_dependencies: ValidationDependencies = (
            self._expectation.get_validation_dependencies()
        )

        metric_name: str
        metric_diagnostics_list: List[ExpectationMetricDiagnostics] = [
            ExpectationMetricDiagnostics(
                name=metric_name,
                has_question_renderer=False,
            )
            for metric_name in validation_dependencies.get_metric_names()
        ]

        return metric_diagnostics_list

    @classmethod
    def _get_test_results(  # noqa: PLR0913 # FIXME CoP
        cls,
        expectation_type: str,
        test_data_cases: List[ExpectationTestDataCases],
        execution_engine_diagnostics: ExpectationExecutionEngineDiagnostics,
        raise_exceptions_for_backends: bool = False,
        ignore_suppress: bool = False,
        ignore_only_for: bool = False,
        debug_logger: Optional[logging.Logger] = None,
        only_consider_these_backends: Optional[List[str]] = None,
        context: Optional[AbstractDataContext] = None,
    ) -> List[ExpectationTestDiagnostics]:
        """Generate test results. This is an internal method for run_diagnostics."""

        if debug_logger is not None:
            _debug = lambda x: debug_logger.debug(  # noqa: E731 # FIXME CoP
                f"(_get_test_results) {x}"
            )
            _error = lambda x: debug_logger.error(  # noqa: E731 # FIXME CoP
                f"(_get_test_results) {x}"
            )
        else:
            _debug = lambda x: x  # noqa: E731 # FIXME CoP
            _error = lambda x: x  # noqa: E731 # FIXME CoP
        _debug("Starting")

        test_results = []

        exp_tests = generate_expectation_tests(
            expectation_type=expectation_type,
            test_data_cases=test_data_cases,
            execution_engine_diagnostics=execution_engine_diagnostics,
            raise_exceptions_for_backends=raise_exceptions_for_backends,
            ignore_suppress=ignore_suppress,
            ignore_only_for=ignore_only_for,
            debug_logger=debug_logger,
            only_consider_these_backends=only_consider_these_backends,
            context=context,
        )

        error_diagnostics: Optional[ExpectationErrorDiagnostics]
        backend_test_times = defaultdict(list)
        for exp_test in exp_tests:
            if exp_test["test"] is None:
                _debug(f"validator_with_data failure for {exp_test['backend']}--{expectation_type}")

                error_diagnostics = ExpectationErrorDiagnostics(
                    error_msg=exp_test["error"],
                    stack_trace="",
                    test_title="all",
                    test_backend=exp_test["backend"],
                )

                test_results.append(
                    ExpectationTestDiagnostics(
                        test_title="all",
                        backend=exp_test["backend"],
                        test_passed=False,
                        include_in_gallery=False,
                        validation_result=None,
                        error_diagnostics=error_diagnostics,
                    )
                )
                continue

            exp_combined_test_name = (
                f"{exp_test['backend']}--{exp_test['test']['title']}--{expectation_type}"
            )
            _debug(f"Starting {exp_combined_test_name}")
            _start = time.time()
            validation_result, error_message, stack_trace = evaluate_json_test_v3_api(
                validator=exp_test["validator_with_data"],
                expectation_type=exp_test["expectation_type"],
                test=exp_test["test"],
                raise_exception=False,
                debug_logger=debug_logger,
            )
            _end = time.time()
            _duration = _end - _start
            backend_test_times[exp_test["backend"]].append(_duration)
            _debug(
                f"Took {_duration} seconds to evaluate_json_test_v3_api for {exp_combined_test_name}"  # noqa: E501 # FIXME CoP
            )
            if error_message is None:
                _debug(f"PASSED {exp_combined_test_name}")
                test_passed = True
                error_diagnostics = None
            else:
                _error(f"{error_message!r} for {exp_combined_test_name}")
                print(f"{stack_trace[0]}")
                error_diagnostics = ExpectationErrorDiagnostics(
                    error_msg=error_message,
                    stack_trace=stack_trace,
                    test_title=exp_test["test"]["title"],
                    test_backend=exp_test["backend"],
                )
                test_passed = False

            if validation_result:
                # The ExpectationTestDiagnostics instance will error when calling it's to_dict()
                # method (AttributeError: 'ExpectationConfiguration' object has no attribute 'raw_kwargs')  # noqa: E501 # FIXME CoP
                validation_result.expectation_config.raw_kwargs = (
                    validation_result.expectation_config._raw_kwargs
                )

            test_results.append(
                ExpectationTestDiagnostics(
                    test_title=exp_test["test"]["title"],
                    backend=exp_test["backend"],
                    test_passed=test_passed,
                    include_in_gallery=exp_test["test"]["include_in_gallery"],
                    validation_result=validation_result,
                    error_diagnostics=error_diagnostics,
                )
            )

        for backend_name, test_times in sorted(backend_test_times.items()):
            _debug(
                f"Took {sum(test_times)} seconds to run {len(test_times)} tests {backend_name}--{expectation_type}"  # noqa: E501 # FIXME CoP
            )

        return test_results

    def _get_renderer_diagnostics(
        self,
        expectation_type: str,
        test_diagnostics: List[ExpectationTestDiagnostics],
        registered_renderers: List[str],
        standard_renderers: Optional[
            List[Union[str, LegacyRendererType, LegacyDiagnosticRendererType]]
        ] = None,
    ) -> List[ExpectationRendererDiagnostics]:
        """Generate Renderer diagnostics for this Expectation, based primarily on a list of ExpectationTestDiagnostics."""  # noqa: E501 # FIXME CoP

        if not standard_renderers:
            standard_renderers = [
                LegacyRendererType.ANSWER,
                LegacyDiagnosticRendererType.UNEXPECTED_STATEMENT,
                LegacyDiagnosticRendererType.OBSERVED_VALUE,
                LegacyDiagnosticRendererType.STATUS_ICON,
                LegacyDiagnosticRendererType.UNEXPECTED_TABLE,
                LegacyRendererType.PRESCRIPTIVE,
                LegacyRendererType.QUESTION,
            ]

        supported_renderers = self._get_registered_renderers(
            expectation_type=expectation_type,
            registered_renderers=registered_renderers,  # type: ignore[arg-type] # FIXME CoP
        )

        renderer_diagnostic_list = []
        for renderer_name in set(standard_renderers).union(set(supported_renderers)):
            samples = []
            if renderer_name in supported_renderers:
                _, renderer = registered_renderers[expectation_type][renderer_name]  # type: ignore[call-overload] # FIXME CoP

                for test_diagnostic in test_diagnostics:
                    test_title = test_diagnostic["test_title"]

                    try:
                        rendered_result = renderer(
                            configuration=test_diagnostic["validation_result"][
                                "expectation_config"
                            ],
                            result=test_diagnostic["validation_result"],
                        )
                        rendered_result_str = self._get_rendered_result_as_string(rendered_result)

                    except Exception as e:
                        new_sample = RendererTestDiagnostics(
                            test_title=test_title,
                            renderered_str=None,
                            rendered_successfully=False,
                            error_message=str(e),
                            stack_trace=traceback.format_exc(),
                        )

                    else:
                        new_sample = RendererTestDiagnostics(
                            test_title=test_title,
                            renderered_str=rendered_result_str,
                            rendered_successfully=True,
                        )

                    finally:
                        samples.append(new_sample)

            new_renderer_diagnostics = ExpectationRendererDiagnostics(
                name=renderer_name,
                is_supported=renderer_name in supported_renderers,
                is_standard=renderer_name in standard_renderers,
                samples=samples,
            )
            renderer_diagnostic_list.append(new_renderer_diagnostics)

        # Sort to enforce consistency for testing
        renderer_diagnostic_list.sort(key=lambda x: x.name)

        return renderer_diagnostic_list

    @staticmethod
    def _get_registered_renderers(
        expectation_type: str,
        registered_renderers: dict,
    ) -> List[str]:
        """Get a list of supported renderers for this Expectation, in sorted order."""
        supported_renderers = list(registered_renderers[expectation_type].keys())
        supported_renderers.sort()
        return supported_renderers

    def _get_rendered_result_as_string(  # noqa: C901, PLR0912 # FIXME CoP
        self, rendered_result
    ) -> str:
        """Convenience method to get rendered results as strings."""

        result: str = ""

        if isinstance(rendered_result, str):
            result = rendered_result

        elif isinstance(rendered_result, list):
            sub_result_list = []
            for sub_result in rendered_result:
                res = self._get_rendered_result_as_string(sub_result)
                if res is not None:
                    sub_result_list.append(res)

            result = "\n".join(sub_result_list)

        elif isinstance(rendered_result, (CollapseContent, RenderedStringTemplateContent)):
            result = rendered_result.__str__()

        elif isinstance(rendered_result, RenderedAtomicContent):
            result = f"(RenderedAtomicContent) {rendered_result.to_json_dict()!r}"

        elif isinstance(rendered_result, RenderedContentBlockContainer):
            result = "(RenderedContentBlockContainer) " + repr(rendered_result.to_json_dict())

        elif isinstance(rendered_result, RenderedTableContent):
            result = f"(RenderedTableContent) {rendered_result.to_json_dict()!r}"

        elif isinstance(rendered_result, RenderedGraphContent):
            result = f"(RenderedGraphContent) {rendered_result.to_json_dict()!r}"

        elif isinstance(rendered_result, ValueListContent):
            result = f"(ValueListContent) {rendered_result.to_json_dict()!r}"

        elif isinstance(rendered_result, dict):
            result = f"(dict) {rendered_result!r}"

        elif isinstance(rendered_result, int):
            result = repr(rendered_result)

        elif rendered_result is None:
            result = ""

        else:
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"Expectation._get_rendered_result_as_string can't render type {type(rendered_result)} as a string."  # noqa: E501 # FIXME CoP
            )

        if "inf" in result:
            result = ""
        return result



import pathlib
import re
from typing import Final, List, Union

import pytest

from great_expectations.alias_types import PathStr
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.partitioners import (
    ColumnPartitionerDaily,
    ColumnPartitionerMonthly,
    ColumnPartitionerYearly,
    FileNamePartitionerDaily,
    FileNamePartitionerMonthly,
    FileNamePartitionerPath,
    FileNamePartitionerYearly,
)
from great_expectations.datasource.fluent import Datasource
from great_expectations.datasource.fluent.data_asset.path.file_asset import (
    AmbiguousPathError,
    PathNotFoundError,
    RegexMissingRequiredGroupsError,
    RegexUnknownGroupsError,
)
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import (
    CSVAsset as PandasCSVAsset,
)
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import (
    ExcelAsset,
    FWFAsset,
)
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import (
    JSONAsset as PandasJSONAsset,
)
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import (
    ORCAsset as PandasORCAsset,
)
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import (
    ParquetAsset as PandasParquetAsset,
)
from great_expectations.datasource.fluent.data_asset.path.path_data_asset import (
    PathDataAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import (
    CSVAsset as SparkCSVAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import (
    DirectoryCSVAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.delta_asset import (
    DeltaAsset,
    DirectoryDeltaAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.json_asset import (
    DirectoryJSONAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.json_asset import (
    JSONAsset as SparkJSONAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.orc_asset import (
    DirectoryORCAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.orc_asset import (
    ORCAsset as SparkORCAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.parquet_asset import (
    DirectoryParquetAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.parquet_asset import (
    ParquetAsset as SparkParquetAsset,
)
from great_expectations.datasource.fluent.data_asset.path.spark.text_asset import (
    DirectoryTextAsset,
    TextAsset,
)
from great_expectations.datasource.fluent.data_connector import FilePathDataConnector
from great_expectations.datasource.fluent.pandas_filesystem_datasource import (
    PandasFilesystemDatasource,
)


@pytest.fixture
def validated_pandas_filesystem_datasource(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    years = ["2018", "2019", "2020"]
    all_files: List[str] = [
        file_name.stem
        for file_name in list(pathlib.Path(pandas_filesystem_datasource.base_directory).iterdir())
    ]
    # assert there are 12 files for each year
    for year in years:
        files_for_year = [
            file_name
            for file_name in all_files
            if file_name.find(f"yellow_tripdata_sample_{year}") == 0
        ]
        assert len(files_for_year) == 12
    return pandas_filesystem_datasource


@pytest.mark.filesystem
@pytest.mark.parametrize(
    "batching_regex",
    [
        pytest.param(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv", id="String"),
        pytest.param(
            re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"),
            id="re.Pattern",
        ),
    ],
)
def test_get_batch_identifiers_list__sort_ascending(
    validated_pandas_filesystem_datasource: PandasFilesystemDatasource,
    batching_regex: Union[str, re.Pattern],
):
    """Verify that get_batch_identifiers_list respects a partitioner's ascending sort order.

    NOTE: we just happen to be using pandas as the concrete class.
    """
    asset = validated_pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    batch_definition = asset.add_batch_definition_monthly(
        name="foo", sort_ascending=True, regex=batching_regex
    )
    batch_request = batch_definition.build_batch_request()

    batch_identifiers_list = asset.get_batch_identifiers_list(batch_request)

    expected_years = ["2018"] * 12 + ["2019"] * 12 + ["2020"] * 12
    expected_months = [format(m, "02d") for m in range(1, 13)] * 3

    assert (len(batch_identifiers_list)) == 36
    for i, batch_identifiers in enumerate(batch_identifiers_list):
        assert batch_identifiers["year"] == str(expected_years[i])
        assert batch_identifiers["month"] == str(expected_months[i])


@pytest.mark.filesystem
@pytest.mark.parametrize(
    "batching_regex",
    [
        pytest.param(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv", id="String"),
        pytest.param(
            re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"),
            id="re.Pattern",
        ),
    ],
)
def test_get_batch_identifiers_list__sort_descending(
    validated_pandas_filesystem_datasource: PandasFilesystemDatasource,
    batching_regex: Union[str, re.Pattern],
):
    """Verify that get_batch_identifiers_list respects a partitioner's descending sort order.

    NOTE: we just happen to be using pandas as the concrete class.
    """
    asset = validated_pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    batch_definition = asset.add_batch_definition_monthly(
        name="foo", regex=batching_regex, sort_ascending=False
    )
    batch_request = batch_definition.build_batch_request()

    batch_identifiers_list = asset.get_batch_identifiers_list(batch_request)

    expected_years = list(reversed(["2018"] * 12 + ["2019"] * 12 + ["2020"] * 12))
    expected_months = list(reversed([format(m, "02d") for m in range(1, 13)] * 3))

    assert (len(batch_identifiers_list)) == 36
    for i, batch_identifiers in enumerate(batch_identifiers_list):
        assert batch_identifiers["year"] == str(expected_years[i])
        assert batch_identifiers["month"] == str(expected_months[i])


@pytest.fixture
def datasource(mocker):
    # the API required by these tests is not specific to Spark or Pandas
    return mocker.Mock(spec=Datasource)


@pytest.fixture
def file_path_data_connector(mocker):
    return mocker.Mock(spec=FilePathDataConnector)


@pytest.fixture
def asset(request, datasource, file_path_data_connector) -> PathDataAsset:
    asset = request.param
    # since we're parametrizing these tests multiple ways, this object
    # will likely be reused, so we make sure the state we care about is reset:
    asset.batch_definitions = []
    asset._datasource = datasource  # same pattern Datasource uses to init Asset
    asset._data_connector = file_path_data_connector
    return asset


PATH_NAME = "data_2022-01.csv"


def _path_asset_parameters():
    return [
        # Spark Assets
        pytest.param(SparkCSVAsset(name="test_asset"), id="Spark CSV Asset"),
        pytest.param(SparkParquetAsset(name="test_asset"), id="Spark Parquet Asset"),
        pytest.param(SparkORCAsset(name="test_asset"), id="Spark ORC Asset"),
        pytest.param(SparkJSONAsset(name="test_asset"), id="Spark JSON Asset"),
        pytest.param(TextAsset(name="test_asset"), id="Spark Text Asset"),
        pytest.param(DeltaAsset(name="test_asset"), id="Spark Delta Asset"),
        # Pandas Assets
        pytest.param(PandasCSVAsset(name="test_asset"), id="Pandas CSV Asset"),
        pytest.param(ExcelAsset(name="test_asset"), id="Pandas Excel Asset"),
        pytest.param(PandasJSONAsset(name="test_asset"), id="Pandas JSON Asset"),
        pytest.param(PandasORCAsset(name="test_asset"), id="Pandas ORC Asset"),
        pytest.param(PandasParquetAsset(name="test_asset"), id="Pandas Parquet Asset"),
        pytest.param(FWFAsset(name="test_asset"), id="Pandas FWF Asset"),
    ]


@pytest.mark.unit
@pytest.mark.parametrize(
    "path",
    [
        pytest.param(PATH_NAME, id="String Path"),
        pytest.param(pathlib.Path(PATH_NAME), id="Pathlib Path"),
    ],
)
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
def test_add_batch_definition_fluent_file_path__add_batch_definition_path_success(
    datasource,
    asset,
    path: PathStr,
    file_path_data_connector,
):
    # arrange
    name = "batch_def_name"
    expected_regex = re.compile(f"{path}$")
    expected_batch_definition = BatchDefinition(
        name=name, partitioner=FileNamePartitionerPath(regex=expected_regex)
    )
    assert isinstance(expected_batch_definition.partitioner, FileNamePartitionerPath)
    datasource.add_batch_definition.return_value = expected_batch_definition
    file_path_data_connector.get_matched_data_references.return_value = [PATH_NAME]

    # act
    batch_definition = asset.add_batch_definition_path(name=name, path=path)

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)
    file_path_data_connector.get_matched_data_references.assert_called_once_with(
        regex=expected_regex
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "path",
    [
        pytest.param(PATH_NAME, id="String Path"),
        pytest.param(pathlib.Path(PATH_NAME), id="Pathlib Path"),
    ],
)
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
def test_add_batch_definition_fluent_file_path__add_batch_definition_path_fails_if_no_file_is_found(
    datasource, asset, path: PathStr, file_path_data_connector
):
    # arrange
    name = "batch_def_name"
    expected_regex = re.compile(f"{path}$")

    file_path_data_connector.get_matched_data_references.return_value = []

    # act
    with pytest.raises(PathNotFoundError) as error:
        asset.add_batch_definition_path(name=name, path=path)

        # assert -- we need to still be inside context manager to access this instance attribute
        assert error.path == path  # type: ignore[attr-defined]  # pytest.raises obscures type

    file_path_data_connector.get_matched_data_references.assert_called_once_with(
        regex=expected_regex
    )
    datasource.add_batch_definition.assert_not_called()


@pytest.mark.unit
@pytest.mark.parametrize(
    "path",
    [
        pytest.param(PATH_NAME, id="String Path"),
        pytest.param(pathlib.Path(PATH_NAME), id="Pathlib Path"),
    ],
)
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
def test_add_batch_definition_fluent_file_path__add_batch_definition_path_fails_if_multiple_files_are_found(  # noqa: E501 # FIXME CoP
    datasource, asset, path: PathStr, file_path_data_connector
):
    """This edge case occurs if a user doesn't actually provide a path, but
    instead a regex with multiple matches."""
    # arrange
    name = "batch_def_name"
    expected_regex = re.compile(f"{path}$")
    file_path_data_connector.get_matched_data_references.return_value = [
        "data_reference_one",
        "data_reference_two",
    ]

    # act
    with pytest.raises(AmbiguousPathError) as error:
        asset.add_batch_definition_path(name=name, path=path)

        # assert -- we need to still be inside context manager to access this instance attribute
        assert error.path == path  # type: ignore[attr-defined]  # pytest.raises obscures type

    file_path_data_connector.get_matched_data_references.assert_called_once_with(
        regex=expected_regex
    )
    datasource.add_batch_definition.assert_not_called()


@pytest.mark.unit
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize(
    "batching_regex",
    [
        pytest.param(r"data_(?P<year>\d{4}).csv", id="String"),
        pytest.param(re.compile(r"data_(?P<year>\d{4}).csv"), id="re.Pattern"),
    ],
)
def test_add_batch_definition_fluent_file_path__add_batch_definition_yearly_success(
    datasource, asset, sort, batching_regex
):
    # arrange
    name = "batch_def_name"
    expected_batch_definition = BatchDefinition(
        name=name,
        partitioner=FileNamePartitionerYearly(regex=batching_regex, sort_ascending=sort),
    )
    assert isinstance(expected_batch_definition.partitioner, FileNamePartitionerYearly)
    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_yearly(
        name=name, regex=batching_regex, sort_ascending=sort
    )

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)


@pytest.mark.unit
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize(
    "batching_regex",
    [
        pytest.param(r"data_2024.csv", id="String"),
        pytest.param(re.compile(r"data_2024.csv"), id="re.Pattern"),
    ],
)
def test_add_batch_definition_fluent_file_path__add_batch_definition_yearly_fails_if_required_group_is_missing(  # noqa: E501 # FIXME CoP
    datasource, asset, sort, batching_regex
):
    # arrange
    name = "batch_def_name"

    # act
    with pytest.raises(RegexMissingRequiredGroupsError) as error:
        asset.add_batch_definition_yearly(name=name, regex=batching_regex, sort_ascending=sort)

        # assert -- we need to still be inside context manager to access this instance attribute
        assert error.missing_groups == {"year"}

    datasource.add_batch_definition.assert_not_called()


@pytest.mark.unit
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize(
    "batching_regex",
    [
        pytest.param(r"data_(?P<year>\d{4})-(?P<foo>\d{4}).csv", id="String"),
        pytest.param(re.compile(r"data_(?P<year>\d{4})-(?P<foo>\d{4}).csv"), id="re.Pattern"),
    ],
)
def test_add_batch_definition_fluent_file_path__add_batch_definition_yearly_fails_if_unknown_groups_are_found(  # noqa: E501 # FIXME CoP
    datasource, asset, sort, batching_regex
):
    # arrange
    name = "batch_def_name"

    # act
    with pytest.raises(RegexUnknownGroupsError) as error:
        asset.add_batch_definition_yearly(name=name, regex=batching_regex, sort_ascending=sort)

        # assert -- we need to still be inside context manager to access this instance attribute
        assert error.unknown_groups == {"foo"}

    datasource.add_batch_definition.assert_not_called()


@pytest.mark.unit
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize(
    "batching_regex",
    [
        pytest.param(r"data_(?P<year>\d{4})-(?P<month>\d{2}).csv", id="String"),
        pytest.param(re.compile(r"data_(?P<year>\d{4})-(?P<month>\d{2}).csv"), id="re.Pattern"),
    ],
)
def test_add_batch_definition_fluent_file_path__add_batch_definition_monthly_success(
    datasource, asset, sort, batching_regex
):
    # arrange
    name = "batch_def_name"
    expected_batch_definition = BatchDefinition(
        name=name,
        partitioner=FileNamePartitionerMonthly(regex=batching_regex, sort_ascending=sort),
    )
    assert isinstance(expected_batch_definition.partitioner, FileNamePartitionerMonthly)

    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_monthly(
        name=name, regex=batching_regex, sort_ascending=sort
    )

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)


@pytest.mark.unit
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize(
    "batching_regex",
    [
        pytest.param(r"data_(?P<year>\d{4})-(?P<month>\d{2}).csv", id="String"),
        pytest.param(re.compile(r"data_(?P<year>\d{4})-(?P<month>\d{2}).csv"), id="re.Pattern"),
    ],
)
def test_add_batch_definition_fluent_file_path__add_batch_definition_monthly_fails_if_required_group_is_missing(  # noqa: E501 # FIXME CoP
    datasource, asset, sort, batching_regex
):
    # arrange
    name = "batch_def_name"
    batching_regex = re.compile(r"data_2024-01.csv")

    # act
    with pytest.raises(RegexMissingRequiredGroupsError) as error:
        asset.add_batch_definition_monthly(name=name, regex=batching_regex, sort_ascending=sort)

        # assert -- we need to still be inside context manager to access this instance attribute
        assert error.missing_groups == {"year", "month"}

    datasource.add_batch_definition.assert_not_called()


@pytest.mark.unit
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
@pytest.mark.parametrize("sort", [True, False])
def test_add_batch_definition_fluent_file_path__add_batch_definition_monthly_fails_if_unknown_groups_are_found(  # noqa: E501 # FIXME CoP
    datasource, asset, sort
):
    # arrange
    name = "batch_def_name"
    batching_regex = re.compile(r"data_(?P<year>\d{4})-(?P<month>\d{2})-(?P<foo>\d{4}).csv")

    # act
    with pytest.raises(RegexUnknownGroupsError) as error:
        asset.add_batch_definition_monthly(name=name, regex=batching_regex, sort_ascending=sort)

        # assert -- we need to still be inside context manager to access this instance attribute
        assert error.unknown_groups == {"foo"}

    datasource.add_batch_definition.assert_not_called()


@pytest.mark.unit
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
@pytest.mark.parametrize("sort", [True, False])
def test_add_batch_definition_fluent_file_path__add_batch_definition_daily_success(
    datasource, asset, sort
):
    # arrange
    name = "batch_def_name"
    batching_regex = re.compile(r"data_(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2}).csv")
    expected_batch_definition = BatchDefinition(
        name=name,
        partitioner=FileNamePartitionerDaily(regex=batching_regex, sort_ascending=sort),
    )
    assert isinstance(expected_batch_definition.partitioner, FileNamePartitionerDaily)

    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_daily(
        name=name, regex=batching_regex, sort_ascending=sort
    )

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)


@pytest.mark.unit
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize(
    "batching_regex",
    [
        pytest.param(r"data_2024.csv", id="String"),
        pytest.param(re.compile(r"data_2024.csv"), id="re.Pattern"),
    ],
)
def test_add_batch_definition_fluent_file_path__add_batch_definition_daily_fails_if_required_group_is_missing(  # noqa: E501 # FIXME CoP
    datasource, asset, sort, batching_regex
):
    # arrange
    name = "batch_def_name"

    # act
    with pytest.raises(RegexMissingRequiredGroupsError) as error:
        asset.add_batch_definition_daily(name=name, regex=batching_regex, sort_ascending=sort)

        # assert -- we need to still be inside context manager to access this instance attribute
        assert error.missing_groups == {"year", "month", "day"}

    datasource.add_batch_definition.assert_not_called()


@pytest.mark.unit
@pytest.mark.parametrize("asset", _path_asset_parameters(), indirect=["asset"])
@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize(
    "batching_regex",
    [
        pytest.param(
            r"data_(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})-(?P<foo>\d{4}).csv", id="String"
        ),
        pytest.param(
            re.compile(r"data_(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})-(?P<foo>\d{4}).csv"),
            id="re.Pattern",
        ),
    ],
)
def test_add_batch_definition_fluent_file_path__add_batch_definition_daily_fails_if_unknown_groups_are_found(  # noqa: E501 # FIXME CoP
    datasource, asset, sort, batching_regex
):
    # arrange
    name = "batch_def_name"
    batching_regex = re.compile(
        r"data_(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})-(?P<foo>\d{4}).csv"
    )

    # act
    with pytest.raises(RegexUnknownGroupsError) as error:
        asset.add_batch_definition_daily(name=name, regex=batching_regex, sort_ascending=sort)

        # assert -- we need to still be inside context manager to access this instance attribute
        assert error.unknown_groups == {"foo"}

    datasource.add_batch_definition.assert_not_called()


DATA_DIRECTORY: Final[pathlib.Path] = pathlib.Path("/data/my_directory")


def _directory_asset_parameters():
    return [
        pytest.param(
            DirectoryCSVAsset(name="test_asset", data_directory=DATA_DIRECTORY), id="CSV Asset"
        ),
        pytest.param(
            DirectoryParquetAsset(name="test_asset", data_directory=DATA_DIRECTORY),
            id="Parquet Asset",
        ),
        pytest.param(
            DirectoryORCAsset(name="test_asset", data_directory=DATA_DIRECTORY), id="ORC Asset"
        ),
        pytest.param(
            DirectoryJSONAsset(name="test_asset", data_directory=DATA_DIRECTORY), id="JSON Asset"
        ),
        pytest.param(
            DirectoryTextAsset(name="test_asset", data_directory=DATA_DIRECTORY), id="Text Asset"
        ),
        pytest.param(
            DirectoryDeltaAsset(name="test_asset", data_directory=DATA_DIRECTORY), id="Delta Asset"
        ),
    ]


@pytest.mark.unit
@pytest.mark.parametrize("asset", _directory_asset_parameters(), indirect=["asset"])
def test_add_batch_definition_whole_directory_success(
    datasource,
    asset,
):
    # arrange
    name = "batch_def_name"
    expected_batch_definition = BatchDefinition(name=name, partitioner=None)
    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_whole_directory(name=name)

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)


@pytest.mark.unit
@pytest.mark.parametrize("asset", _directory_asset_parameters(), indirect=["asset"])
def test_add_batch_definition_daily_success(datasource, asset):
    # arrange
    name = "batch_def_name"
    column = "foo"
    expected_batch_definition = BatchDefinition(
        name=name,
        partitioner=ColumnPartitionerDaily(column_name=column),
    )
    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_daily(name=name, column=column)

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)


@pytest.mark.unit
@pytest.mark.parametrize("asset", _directory_asset_parameters(), indirect=["asset"])
def test_add_batch_definition_monthly_success(
    datasource,
    asset,
):
    # arrange
    name = "batch_def_name"
    column = "foo"
    expected_batch_definition = BatchDefinition(
        name=name, partitioner=ColumnPartitionerMonthly(column_name=column)
    )
    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_monthly(name=name, column=column)

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)


@pytest.mark.unit
@pytest.mark.parametrize("asset", _directory_asset_parameters(), indirect=["asset"])
def test_add_batch_definition_yearly_success(
    datasource,
    asset,
):
    # arrange
    name = "batch_def_name"
    column = "foo"
    expected_batch_definition = BatchDefinition(
        name=name, partitioner=ColumnPartitionerYearly(column_name=column)
    )
    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_yearly(name=name, column=column)

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)



from __future__ import annotations

import warnings
from typing import Callable, Iterator

import pandas as pd

from great_expectations.compatibility.not_imported import (
    is_version_greater_or_equal,
    is_version_less_than,
)
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.warnings import (
    warn_pandas_less_than_2_0_and_sqlalchemy_greater_than_or_equal_2_0,
)


def execute_pandas_reader_fn(
    reader_fn: Callable, reader_options: dict
) -> pd.DataFrame | list[pd.DataFrame]:
    """Suppress warnings while executing the pandas reader functions.

    If pandas version is below 2.0 and sqlalchemy installed then we suppress
    the sqlalchemy 2.0 warning and raise our own warning. pandas does not
    support sqlalchemy 2.0 until version 2.0 (see https://pandas.pydata.org/docs/dev/whatsnew/v2.0.0.html#other-enhancements)

    Args:
        reader_fn: Reader function to execute.
        reader_options: Options to pass to reader function.

    Returns:
        dataframe or list of dataframes
    """
    if is_version_less_than(pd.__version__, "2.0.0"):
        if sa and is_version_greater_or_equal(sa.__version__, "2.0.0"):
            warn_pandas_less_than_2_0_and_sqlalchemy_greater_than_or_equal_2_0()
        with warnings.catch_warnings():
            # Note that RemovedIn20Warning is the warning class that we see from sqlalchemy
            # but using the base class here since sqlalchemy is an optional dependency and this
            # warning type only exists in sqlalchemy < 2.0.
            warnings.filterwarnings(action="ignore", category=DeprecationWarning)
            reader_fn_result: pd.DataFrame | list[pd.DataFrame] = reader_fn(**reader_options)
    else:
        reader_fn_result = reader_fn(**reader_options)
    return reader_fn_result


def pandas_read_sql(sql, con, **kwargs) -> pd.DataFrame | Iterator[pd.DataFrame]:
    """Suppress deprecation warnings while executing the pandas read_sql function.

    Note this only passes params straight to pandas read_sql method, please
    see the pandas documentation
    (currently https://pandas.pydata.org/docs/reference/api/pandas.read_sql.html)
    for more information on this method.

    If pandas version is below 2.0 and sqlalchemy installed then we suppress
    the sqlalchemy 2.0 warning and raise our own warning. pandas does not
    support sqlalchemy 2.0 until version 2.0 (see https://pandas.pydata.org/docs/dev/whatsnew/v2.0.0.html#other-enhancements)

    Args:
        sql: str or SQLAlchemy Selectable (select or text object)
        con: SQLAlchemy connectable, str, or sqlite3 connection
        **kwargs: Other keyword arguments, not enumerated here since they differ
            between pandas versions.

    Returns:
        dataframe
    """
    if is_version_less_than(pd.__version__, "2.0.0"):
        if sa and is_version_greater_or_equal(sa.__version__, "2.0.0"):
            warn_pandas_less_than_2_0_and_sqlalchemy_greater_than_or_equal_2_0()
        with warnings.catch_warnings():
            # Note that RemovedIn20Warning is the warning class that we see from sqlalchemy
            # but using the base class here since sqlalchemy is an optional dependency and this
            # warning type only exists in sqlalchemy < 2.0.
            warnings.filterwarnings(action="ignore", category=DeprecationWarning)
            return_value = pd.read_sql(sql=sql, con=con, **kwargs)
    else:
        if not sql.supports_execution:
            sql = sa.select(sa.text("*")).select_from(sql)
        return_value = pd.read_sql(sql=sql, con=con, **kwargs)
    return return_value



from __future__ import annotations

import logging
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
)

from marshmallow import ValidationError as MarshmallowValidationError
from typing_extensions import TypedDict

import great_expectations.exceptions as gx_exceptions
from great_expectations.analytics import submit as submit_analytics_event
from great_expectations.analytics.events import DomainObjectAllDeserializationEvent
from great_expectations.compatibility.pydantic import ValidationError as PydanticValidationError
from great_expectations.core.data_context_key import DataContextKey
from great_expectations.data_context.store.gx_cloud_store_backend import (
    GXCloudStoreBackend,
)
from great_expectations.data_context.store.store_backend import StoreBackend
from great_expectations.data_context.types.resource_identifiers import (
    ConfigurationIdentifier,
    GXCloudIdentifier,
)
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.exceptions import (
    ClassInstantiationError,
    DataContextError,
    StoreBackendError,
)
from great_expectations.util import load_class, verify_dynamic_loading_support

if TYPE_CHECKING:
    # min version of typing_extension missing `NotRequired`, so it can't be imported at runtime
    from typing_extensions import NotRequired

    from great_expectations.core.configuration import AbstractConfig

logger = logging.getLogger(__name__)


class StoreConfigTypedDict(TypedDict):
    # NOTE: TypeDict values may be incomplete, update as needed
    class_name: str
    module_name: NotRequired[str]
    store_backend: dict


class DataDocsSiteConfigTypedDict(TypedDict):
    # NOTE: TypeDict values may be incomplete, update as needed
    class_name: str
    module_name: NotRequired[str]
    store_backend: dict
    site_index_builder: dict


class Store:
    """A store is responsible for reading and writing Great Expectations objects
    to appropriate backends. It provides a generic API that the DataContext can
    use independently of any particular ORM and backend.

    An implementation of a store will generally need to define the following:
      - serialize
      - deserialize
      - _key_class (class of expected key type)

    All keys must have a to_tuple() method.
    """

    _key_class: ClassVar[Type] = DataContextKey

    def __init__(
        self,
        store_backend: Optional[dict] = None,
        runtime_environment: Optional[dict] = None,
        store_name: str = "no_store_name",
    ) -> None:
        """
        Runtime environment may be necessary to instantiate store backend elements.
        Args:
            store_backend:
            runtime_environment:
            store_name: store name given in the DataContextConfig (via either in-code or yaml configuration)
        """  # noqa: E501 # FIXME CoP
        if store_backend is None:
            store_backend = {"class_name": "InMemoryStoreBackend"}
        self._store_name = store_name
        logger.debug("Building store_backend.")
        module_name = "great_expectations.data_context.store"
        self._store_backend = instantiate_class_from_config(
            config=store_backend,
            runtime_environment=runtime_environment or {},
            config_defaults={
                "module_name": module_name,
                "store_name": self._store_name,
            },
        )
        if not self._store_backend:
            raise ClassInstantiationError(
                module_name=module_name, package_name=None, class_name=store_backend
            )
        if not isinstance(self._store_backend, StoreBackend):
            raise DataContextError(  # noqa: TRY003 # FIXME CoP
                "Invalid StoreBackend configuration: expected a StoreBackend instance."
            )
        self._use_fixed_length_key = self._store_backend.fixed_length_key

    @staticmethod
    def _determine_store_backend_class(store_backend: dict | None) -> type:
        store_backend = store_backend or {}
        store_backend_module_name = store_backend.get(
            "module_name", "great_expectations.data_context.store"
        )
        store_backend_class_name = store_backend.get("class_name", "InMemoryStoreBackend")
        verify_dynamic_loading_support(module_name=store_backend_module_name)
        return load_class(store_backend_class_name, store_backend_module_name)

    @classmethod
    def gx_cloud_response_json_to_object_dict(cls, response_json: Dict) -> Dict:
        """
        This method takes full json response from GX cloud and outputs a dict appropriate for
        deserialization into a GX object
        """
        return response_json

    @classmethod
    def gx_cloud_response_json_to_object_collection(cls, response_json: Dict) -> List[Dict]:
        """
        This method takes full json response from GX cloud and outputs a list of dicts appropriate for
        deserialization into a collection of GX objects
        """  # noqa: E501 # FIXME CoP
        logger.debug(f"GE Cloud Response JSON ->\n{pf(response_json, depth=3)}")
        data = response_json["data"]
        if not isinstance(data, list):
            raise TypeError("GX Cloud did not return a collection of Datasources when expected")  # noqa: TRY003 # FIXME CoP

        return [cls._convert_raw_json_to_object_dict(d) for d in data]

    @staticmethod
    def _convert_raw_json_to_object_dict(data: dict[str, Any]) -> dict[str, Any]:
        """Method to convert data from API to raw object dict

        This SHOULD be used by both gx_cloud_response_json_to_object_collection
        and gx_cloud_response_json_to_object_dict. It is a means of keeping
        response parsing DRY for different response types, e.g. collections
        may be shaped like {"data": [item1, item2, ...]} while single items
        may be shaped like {"data": item}. This allows for pulling out the
        data key and passing it to the appropriate method for conversion.
        """
        raise NotImplementedError

    def _validate_key(self, key: DataContextKey) -> None:
        # STORE_BACKEND_ID_KEY always validated
        if key == StoreBackend.STORE_BACKEND_ID_KEY or isinstance(key, self.key_class):
            return
        else:
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"key must be an instance of {self.key_class.__name__}, not {type(key)}"
            )

    @property
    def cloud_mode(self) -> bool:
        return isinstance(self._store_backend, GXCloudStoreBackend)

    @property
    def store_backend(self) -> StoreBackend:
        return self._store_backend

    @property
    def store_name(self) -> str:
        return self._store_name

    @property
    def store_backend_id(self) -> str:
        """
        Report the store_backend_id of the currently-configured StoreBackend
        Returns:
            store_backend_id which is a UUID(version=4)
        """
        return self._store_backend.store_backend_id

    @property
    def key_class(self) -> Type[DataContextKey]:
        if self.cloud_mode:
            return GXCloudIdentifier
        return self._key_class

    @property
    def store_backend_id_warnings_suppressed(self) -> str:
        """
        Report the store_backend_id of the currently-configured StoreBackend, suppressing warnings for invalid configurations.
        Returns:
            store_backend_id which is a UUID(version=4)
        """  # noqa: E501 # FIXME CoP
        return self._store_backend.store_backend_id_warnings_suppressed

    @property
    def config(self) -> dict:
        raise NotImplementedError

    # noinspection PyMethodMayBeStatic
    def serialize(self, value: Any) -> Any:
        return value

    # noinspection PyMethodMayBeStatic
    def key_to_tuple(self, key: DataContextKey) -> Tuple[str, ...]:
        if self._use_fixed_length_key:
            return key.to_fixed_length_tuple()
        return key.to_tuple()

    def tuple_to_key(self, tuple_: Tuple[str, ...]) -> DataContextKey:
        if tuple_ == StoreBackend.STORE_BACKEND_ID_KEY:
            return StoreBackend.STORE_BACKEND_ID_KEY[0]  # type: ignore[return-value] # FIXME CoP
        if self._use_fixed_length_key:
            return self.key_class.from_fixed_length_tuple(tuple_)
        return self.key_class.from_tuple(tuple_)

    # noinspection PyMethodMayBeStatic
    def deserialize(self, value: Any) -> Any:
        return value

    def get(
        self, key: DataContextKey | GXCloudIdentifier | ConfigurationIdentifier
    ) -> Optional[Any]:
        if key == StoreBackend.STORE_BACKEND_ID_KEY:
            return self._store_backend.get(key)

        if self.cloud_mode:
            self._validate_key(key)
            value = self._store_backend.get(self.key_to_tuple(key))
            # TODO [Robby] MER-285: Handle non-200 http errors
            if value:
                value = self.gx_cloud_response_json_to_object_dict(response_json=value)
        else:
            self._validate_key(key)
            value = self._store_backend.get(self.key_to_tuple(key))

        if value:
            return self.deserialize(value)

        return None

    def get_all(self) -> list[Any]:
        objs = self._store_backend.get_all()
        if self.cloud_mode:
            objs = self.gx_cloud_response_json_to_object_collection(objs)

        deserializable_objs: list[Any] = []
        bad_objs: list[Any] = []
        for obj in objs:
            try:
                deserializable_objs.append(self.deserialize(obj))
            except (
                MarshmallowValidationError,
                PydanticValidationError,
                StoreBackendError,
            ) as e:
                bad_objs.append(obj)
                self.submit_all_deserialization_event(e)
            except Exception as e:
                # For a general error we want to log so we can understand if there
                # is user pain here and then we reraise.
                self.submit_all_deserialization_event(e)
                raise

        if bad_objs:
            prefix = "\n    SKIPPED: "
            skipped = prefix + prefix.join([str(bad) for bad in bad_objs])
            logger.warning(f"Skipping Bad Configs:{skipped}")
        return deserializable_objs

    def submit_all_deserialization_event(self, e: Exception):
        error_type = type(e)
        submit_analytics_event(
            DomainObjectAllDeserializationEvent(
                error_type=f"{error_type.__module__}.{error_type.__qualname__}",
                store_name=type(self).__name__,
            )
        )

    def set(self, key: DataContextKey, value: Any, **kwargs) -> Any:
        if key == StoreBackend.STORE_BACKEND_ID_KEY:
            return self._store_backend.set(key, value, **kwargs)

        self._validate_key(key)
        return self._store_backend.set(self.key_to_tuple(key), self.serialize(value), **kwargs)

    def add(self, key: DataContextKey, value: Any, **kwargs) -> None:
        """
        Essentially `set` but validates that a given key-value pair does not already exist.
        """
        return self._add(key=key, value=value, **kwargs)

    def _add(self, key: DataContextKey, value: Any, **kwargs) -> Any:
        self._validate_key(key)
        output = self._store_backend.add(self.key_to_tuple(key), self.serialize(value), **kwargs)
        if hasattr(value, "id") and hasattr(output, "id"):
            value.id = output.id
        return output

    def update(self, key: DataContextKey, value: Any, **kwargs) -> None:
        """
        Essentially `set` but validates that a given key-value pair does already exist.
        """
        return self._update(key=key, value=value, **kwargs)

    def _update(self, key: DataContextKey, value: Any, **kwargs) -> None:
        self._validate_key(key)
        return self._store_backend.update(self.key_to_tuple(key), self.serialize(value), **kwargs)

    def add_or_update(self, key: DataContextKey, value: Any, **kwargs) -> None | GXCloudIdentifier:
        """
        Conditionally calls `add` or `update` based on the presence of the given key.
        """
        return self._add_or_update(key=key, value=value, **kwargs)

    def _add_or_update(self, key: DataContextKey, value: Any, **kwargs) -> None | GXCloudIdentifier:
        self._validate_key(key)
        return self._store_backend.add_or_update(
            self.key_to_tuple(key), self.serialize(value), **kwargs
        )

    def list_keys(self) -> List[DataContextKey]:
        keys_without_store_backend_id = [
            key
            for key in self._store_backend.list_keys()
            if key != StoreBackend.STORE_BACKEND_ID_KEY
        ]
        return [self.tuple_to_key(key) for key in keys_without_store_backend_id]

    def has_key(self, key: DataContextKey) -> bool:
        if key == StoreBackend.STORE_BACKEND_ID_KEY:
            return self._store_backend.has_key(key)
        else:
            if self._use_fixed_length_key:
                return self._store_backend.has_key(key.to_fixed_length_tuple())
            return self._store_backend.has_key(key.to_tuple())

    def remove_key(self, key):
        return self.store_backend.remove_key(key)

    def _build_key_from_config(self, config: AbstractConfig) -> DataContextKey:
        id: Optional[str] = None
        # Chetan - 20220831 - Explicit fork in logic to cover legacy behavior (particularly around Checkpoints).  # noqa: E501 # FIXME CoP
        if hasattr(config, "id"):
            id = config.id

        name: Optional[str] = None
        if hasattr(config, "name"):
            name = config.name

        return self.store_backend.build_key(name=name, id=id)

    @staticmethod
    def build_store_from_config(
        name: Optional[str] = None,
        config: StoreConfigTypedDict | dict | None = None,
        module_name: str = "great_expectations.data_context.store",
        runtime_environment: Optional[dict] = None,
    ) -> Store:
        if config is None or module_name is None:
            raise gx_exceptions.StoreConfigurationError(  # noqa: TRY003 # FIXME CoP
                "Cannot build a store without both a store_config and a module_name"
            )

        try:
            config_defaults: dict = {
                "store_name": name,
                "module_name": module_name,
            }
            new_store = instantiate_class_from_config(
                config=config,
                runtime_environment=runtime_environment,
                config_defaults=config_defaults,
            )
        except gx_exceptions.DataContextError as e:
            logger.critical(f"Error {e} occurred while attempting to instantiate a store.")
            class_name: str = config["class_name"]
            module_name = config.get("module_name", module_name)
            raise gx_exceptions.ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=class_name,
            ) from e

        return new_store



import pytest

from great_expectations.core.expectation_diagnostics.expectation_doctor import (
    ExpectationDoctor,
)
from great_expectations.core.expectation_diagnostics.expectation_test_data_cases import (
    ExpectationTestDataCases,
)
from great_expectations.core.expectation_diagnostics.supporting_types import (
    AugmentedLibraryMetadata,
    ExpectationDescriptionDiagnostics,
    ExpectationExecutionEngineDiagnostics,
    ExpectationMetricDiagnostics,
)
from great_expectations.expectations.expectation import ColumnMapExpectation
from tests.expectations.fixtures.expect_column_values_to_equal_three import (
    ExpectColumnValuesToEqualThree,
    ExpectColumnValuesToEqualThree__SecondIteration,
    ExpectColumnValuesToEqualThree__ThirdIteration,
)

### Tests for _get_augmented_library_metadata


@pytest.mark.unit
def test__get_augmented_library_metadata_on_a_class_with_no_library_metadata_object():
    expectation = ExpectColumnValuesToEqualThree(column="values")
    doctor = ExpectationDoctor(expectation=expectation)
    augmented_library_metadata = doctor._get_augmented_library_metadata()
    assert augmented_library_metadata == AugmentedLibraryMetadata(
        maturity="CONCEPT_ONLY",
        tags=[],
        contributors=[],
        requirements=[],
        library_metadata_passed_checks=False,
        has_full_test_suite=False,
        manually_reviewed_code=False,
        problems=["No library_metadata attribute found"],
    )


@pytest.mark.unit
def test__get_augmented_library_metadata_on_a_class_with_a_basic_library_metadata_object():
    expectation = ExpectColumnValuesToEqualThree__SecondIteration(column="values")
    doctor = ExpectationDoctor(expectation=expectation)
    augmented_library_metadata = doctor._get_augmented_library_metadata()
    assert augmented_library_metadata == AugmentedLibraryMetadata(
        maturity="EXPERIMENTAL",
        tags=["tag", "other_tag"],
        contributors=["@abegong"],
        requirements=[],
        library_metadata_passed_checks=True,
        has_full_test_suite=False,
        manually_reviewed_code=False,
    )


### Tests for _get_examples


@pytest.mark.unit
def test__get_examples_from_a_class_with_no_examples():
    expectation = ExpectColumnValuesToEqualThree(column="values")
    doctor = ExpectationDoctor(expectation=expectation)
    examples = doctor._get_examples()
    assert examples == []


@pytest.mark.unit
def test__get_examples_from_a_class_with_some_examples():
    expectation = ExpectColumnValuesToEqualThree__SecondIteration(column="values")
    doctor = ExpectationDoctor(expectation=expectation)
    examples = doctor._get_examples()
    assert len(examples) == 1

    first_example = examples[0]
    assert isinstance(first_example, ExpectationTestDataCases)
    assert first_example.data == {"mostly_threes": [3, 3, 3, 3, 3, 3, 2, -1, None, None]}
    assert len(first_example.tests) == 1


@pytest.mark.unit
def test__get_examples_from_a_class_with_return_only_gallery_examples_equals_false():
    expectation = ExpectColumnValuesToEqualThree__SecondIteration(column="values")
    doctor = ExpectationDoctor(expectation=expectation)
    examples = doctor._get_examples(return_only_gallery_examples=False)
    assert len(examples) == 1

    first_example = examples[0]
    assert isinstance(first_example, ExpectationTestDataCases)
    assert first_example.data == {"mostly_threes": [3, 3, 3, 3, 3, 3, 2, -1, None, None]}
    assert len(first_example.tests) == 3


### Tests for _get_description_diagnostics


@pytest.mark.unit
def test__get_description_diagnostics():
    class ExpectColumnValuesToBeAwesome(ColumnMapExpectation):
        """Lo, here is a docstring

        It has more to it.
        """

    expectation = ExpectColumnValuesToBeAwesome(column="values")
    doctor = ExpectationDoctor(expectation=expectation)
    description_diagnostics = doctor._get_description_diagnostics()
    assert description_diagnostics == ExpectationDescriptionDiagnostics(
        camel_name="ExpectColumnValuesToBeAwesome",
        snake_name="expect_column_values_to_be_awesome",
        short_description="Lo, here is a docstring",
        docstring="""Lo, here is a docstring

        It has more to it.
        """,
    )


### Tests for _get_metric_diagnostics_list
@pytest.mark.unit
def test__get_metric_diagnostics_list_on_a_class_without_metrics():
    _config = None
    expectation = ExpectColumnValuesToEqualThree(column="values")
    doctor = ExpectationDoctor(expectation=expectation)
    metric_diagnostics_list = doctor._get_metric_diagnostics_list(expectation_config=_config)
    assert len(metric_diagnostics_list) == 0
    ExpectationMetricDiagnostics(
        name="column_values.something",
        has_question_renderer=False,
    )


@pytest.mark.unit
def test__get_metric_diagnostics_list_on_a_class_with_metrics():
    _config = None
    expectation = ExpectColumnValuesToEqualThree__ThirdIteration(column="values")
    doctor = ExpectationDoctor(expectation=expectation)
    metric_diagnostics_list = doctor._get_metric_diagnostics_list(expectation_config=_config)
    assert len(metric_diagnostics_list) == 0
    ExpectationMetricDiagnostics(
        name="column_values.something",
        has_question_renderer=False,
    )


### Tests for _get_execution_engine_diagnostics
@pytest.mark.skip(
    """
The business logic for _get_execution_engine_diagnostics is just plain wrong.
We should be verifying that all expectations test cases run on any given Execution Engine.
Metrics could be used to make inferences, but they'd never provide comparably compelling evidence.
"""
)
@pytest.mark.unit
def test__get_execution_engine_diagnostics_with_no_metrics_diagnostics():
    assert ExpectColumnValuesToEqualThree__ThirdIteration._get_execution_engine_diagnostics(
        metric_diagnostics_list=[],
        registered_metrics={},
    ) == ExpectationExecutionEngineDiagnostics(
        PandasExecutionEngine=False,
        SqlAlchemyExecutionEngine=False,
        SparkDFExecutionEngine=False,
    )


@pytest.mark.skip(
    """
The business logic for _get_execution_engine_diagnostics is just plain wrong.
We should be verifying that all expectations test cases run on any given Execution Engine.
Metrics could be used to make inferences, but they'd never provide comparably compelling evidence.
"""
)
@pytest.mark.unit
def test__get_execution_engine_diagnostics_with_one_metrics_diagnostics():
    metrics_diagnostics_list = [
        ExpectationMetricDiagnostics(
            name="colum_values.something",
            has_question_renderer=True,
        )
    ]
    registered_metrics = {"colum_values.something": {"providers": ["PandasExecutionEngine"]}}
    assert ExpectColumnValuesToEqualThree__ThirdIteration._get_execution_engine_diagnostics(
        metric_diagnostics_list=metrics_diagnostics_list,
        registered_metrics=registered_metrics,
    ) == ExpectationExecutionEngineDiagnostics(
        PandasExecutionEngine=False,
        SqlAlchemyExecutionEngine=False,
        SparkDFExecutionEngine=False,
    )


### Tests for _get_test_results
@pytest.mark.skip(
    reason="Timeout of 30 seconds reached trying to connect to localhost:8088 (trino port)"
)
@pytest.mark.all_backends
def test__get_test_results():
    test_results = ExpectColumnValuesToEqualThree__ThirdIteration(
        column="values"
    )._get_test_results(
        expectation_type="expect_column_values_to_equal_three",
        test_data_cases=ExpectColumnValuesToEqualThree__ThirdIteration(
            column="values"
        )._get_examples(return_only_gallery_examples=False),
        execution_engine_diagnostics=ExpectationExecutionEngineDiagnostics(
            PandasExecutionEngine=True,
            SqlAlchemyExecutionEngine=False,
            SparkDFExecutionEngine=False,
        ),
    )
    assert len(test_results) == 3
    for result in test_results:
        assert result.test_passed

    test_results = ExpectColumnValuesToEqualThree__ThirdIteration(
        column="values"
    )._get_test_results(
        expectation_type="expect_column_values_to_equal_three",
        test_data_cases=ExpectColumnValuesToEqualThree__ThirdIteration(
            column="values"
        )._get_examples(return_only_gallery_examples=False),
        execution_engine_diagnostics=ExpectationExecutionEngineDiagnostics(
            PandasExecutionEngine=True,
            SqlAlchemyExecutionEngine=True,
            SparkDFExecutionEngine=False,
        ),
    )
    for result in test_results:
        # Abe: 1/1/2022: I'm not sure this is the behavior we want long term. How does backend relate to ExecutionEngine?  # noqa: E501 # FIXME CoP
        if result.backend == "pandas":
            assert result.test_passed is True
        elif result.backend == "sqlite":
            assert result.test_passed is False



from enum import Enum
from typing import Dict, List, Optional, Sequence

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.expectation_validation_result import ExpectationValidationResult
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
    SQL_DATA_SOURCES,
    BigQueryDatasourceTestConfig,
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    MSSQLDatasourceTestConfig,
    MySQLDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

COL_A = "col_a"
COL_B = "col_b"
COL_C = "col_c"


DATA = pd.DataFrame(
    {
        COL_A: [1],
        COL_B: [2],
        COL_C: [3],
    }
)


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_golden_path(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectTableColumnsToMatchSet(column_set=[COL_A, COL_B, COL_C])
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectTableColumnsToMatchSet(column_set=[COL_C, COL_A, COL_B]),
            id="order_doesnt_matter",
        ),
        pytest.param(
            gxe.ExpectTableColumnsToMatchSet(column_set=[COL_A, COL_B], exact_match=False),
            id="allows_subset_for_non_exact_match",
        ),
        pytest.param(
            gxe.ExpectTableColumnsToMatchSet(column_set=[COL_A, COL_B, COL_C], exact_match=False),
            id="allows_all_for_non_exact_match",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch, expectation: gxe.ExpectTableColumnsToMatchSet
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectTableColumnsToMatchSet(column_set=[COL_A, COL_B], exact_match=True),
            id="requires_all_columns_for_exact_match",
        ),
        pytest.param(
            gxe.ExpectTableColumnsToMatchSet(column_set=[COL_A]),
            id="defaults_to_exact_match",
        ),
        pytest.param(
            gxe.ExpectTableColumnsToMatchSet(
                column_set=[COL_A, COL_B, COL_C, "col_d"], exact_match=True
            ),
            id="does_not_allow_extras",
        ),
        pytest.param(
            gxe.ExpectTableColumnsToMatchSet(
                column_set=[COL_A, COL_B, COL_C, "col_d"], exact_match=False
            ),
            id="does_not_allow_extrasfor_non_exact_match",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch, expectation: gxe.ExpectTableColumnsToMatchSet
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(False, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_exact_match_(
    batch_for_datasource: Batch, suite_param_value: str, expected_result: bool
) -> None:
    suite_param_key = "test_expect_table_columns_to_match_set"
    expectation = gxe.ExpectTableColumnsToMatchSet(
        column_set=[COL_A, COL_B],
        exact_match={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


# Case insenstivity tests
# For some of our case insensitive tests we break out Snowflake and Redshift into separate
# tests for the following reasons:
##### Snowflake #####
# In test setup, sqlalchemy's CREATE TABLE will not quote lowercase column names
# but will quote uppercase. Since snowflake stores case insensitive strings
# as uppercase, while all the other databases we are testing stores them as lowercase
# this creates different case insensitive columns. We break out the snowflake tests
# into their own tests.
#
#### Redshift ####
# Redshift by default is case insensitive so trying to match case via quoted identifiers
# fails. There is a global setting you can set on a redshift cluster,
# enable_case_sensitive_identifier, one can apply to make it case sensitive. Our code
# looks like it would handle this successfully but I haven't verified since I need to
# change our redshift CI cluster. There is tracked in GX-1197.

CASE_INSENSITIVE_DATA = pd.DataFrame(
    {
        "column_a": [1],
        "COLUMN_B": [2],
        "CoLuMn_C": [3],
    }
)


def observed_column_names(datasource_type: str) -> List[str]:
    """Returns observed column name base on type in A, B, C order."""
    if datasource_type == "snowflake":
        # Since SQLAlchemy doesn't quote lowercase names in CREATE TABLE and since uppercase is
        # case insenstive in Snowflake, both column_a and column_b will be case insensitive.
        # SQLAlchemy will return case insensitive names in lowercase when inspecting the table.
        return [
            "column_a",
            "column_b",
            "CoLuMn_C",
        ]
    elif datasource_type == "redshift":
        # Redshift is case insensitive by default
        # SQLAlchemy will return case insensitive names in lowercase when inspecting the table.
        return [
            "column_a",
            "column_b",
            "column_c",
        ]
    else:
        # For most datasources we expected the observed column names to match CASE_INSENSITIVE_DATA.
        return [
            "column_a",
            "COLUMN_B",
            "CoLuMn_C",
        ]


SQL_DATA_SOURCES_WITHOUT_SNOWFLAKE_REDSHIFT: Sequence[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    MSSQLDatasourceTestConfig(),
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]


@pytest.mark.unit
def test_sql_data_sources_without_snowflake_redshift() -> None:
    # Verify SQL_DATA_SOURCES_WITHOUT_SNOWFLAKE_REDSHIFT is what is says it is
    assert len(SQL_DATA_SOURCES_WITHOUT_SNOWFLAKE_REDSHIFT) + 2 == len(SQL_DATA_SOURCES)
    assert SnowflakeDatasourceTestConfig() not in SQL_DATA_SOURCES_WITHOUT_SNOWFLAKE_REDSHIFT
    assert RedshiftDatasourceTestConfig() not in SQL_DATA_SOURCES_WITHOUT_SNOWFLAKE_REDSHIFT
    for datasource in SQL_DATA_SOURCES_WITHOUT_SNOWFLAKE_REDSHIFT:
        assert datasource in SQL_DATA_SOURCES


@parameterize_batch_for_data_sources(
    data_source_configs=SQL_DATA_SOURCES,
    data=CASE_INSENSITIVE_DATA,
)
def test_case_insensitive_success(batch_for_datasource: Batch) -> None:
    # Arrange
    expectation = gxe.ExpectTableColumnsToMatchSet(column_set=["COLUMN_A", "column_b", "COLumN_c"])

    # Act
    result = batch_for_datasource.validate(expectation)
    result.render()  # creates rendered content

    # Assert
    assert result.success

    # Assert rendered content is as expected
    observed_values = _extract_observed_state(result)
    assert observed_values == dict(
        zip(
            observed_column_names(batch_for_datasource.datasource.type),
            ["expected", "expected", "expected"],
        )
    )
    expected_values = _extract_expected_state(result)
    assert expected_values == {"COLUMN_A": None, "column_b": None, "COLumN_c": None}


@parameterize_batch_for_data_sources(
    data_source_configs=SQL_DATA_SOURCES,
    data=CASE_INSENSITIVE_DATA,
)
def test_case_insensitive_failure(batch_for_datasource: Batch) -> None:
    # Arrange
    expectation = gxe.ExpectTableColumnsToMatchSet(column_set=["COLUMN_Az", "column_b", "COLumN_c"])

    # Act
    result = batch_for_datasource.validate(expectation)
    result.render()  # creates rendered content

    # Assert
    assert not result.success

    # Assert rendered content is as expected
    observed_values = _extract_observed_state(result)
    assert observed_values == dict(
        zip(
            observed_column_names(batch_for_datasource.datasource.type),
            ["unexpected", "expected", "expected"],
        )
    )
    expected_values = _extract_expected_state(result)
    assert expected_values == {"COLUMN_Az": "missing", "column_b": None, "COLumN_c": None}


def _extract_observed_state(result: ExpectationValidationResult) -> Dict[str, str]:
    """Extracts observed column name to rendered state from validation result"""
    # ExpectationValidationResult is not narrowly typed so we need to a lot of
    # isinstance assertions for our expectation.
    assert isinstance(result.rendered_content, list)
    assert len(result.rendered_content) == 1
    value = result.rendered_content[0].to_json_dict()["value"]
    assert isinstance(value, dict)
    rendered_params = value["params"]
    assert isinstance(rendered_params, dict)
    # We don't use a comprehension because we need to do type assertions with
    # isinstance calls.
    observed_state: Dict[str, str] = {}
    for k, v in rendered_params.items():
        assert isinstance(v, dict)
        if k.startswith("ov__"):
            assert isinstance(v["value"], str)
            assert isinstance(v["render_state"], Enum)
            observed_state[v["value"]] = v["render_state"].value
    return observed_state


def _extract_expected_state(result: ExpectationValidationResult) -> Dict[str, Optional[str]]:
    """Extracts expected column name to rendered state from validation result"""
    # ExpectationValidationResult is not narrowly typed so we need to a lot of
    # isinstance assertions for our expectation.
    assert isinstance(result.rendered_content, list)
    assert len(result.rendered_content) == 1
    value = result.rendered_content[0].to_json_dict()["value"]
    assert isinstance(value, dict)
    rendered_params = value["params"]
    assert isinstance(rendered_params, dict)
    # We don't use a comprehension because we need to do type assertions with
    # isinstance calls.
    expected_state: Dict[str, Optional[str]] = {}
    for k, v in rendered_params.items():
        assert isinstance(v, dict)
        if k.startswith("exp__"):
            assert isinstance(v["value"], str)
            if "render_state" in v and isinstance(v["render_state"], Enum):
                expected_state[v["value"]] = v["render_state"].value
            else:
                expected_state[v["value"]] = None
    return expected_state


@parameterize_batch_for_data_sources(
    data_source_configs=SQL_DATA_SOURCES_WITHOUT_SNOWFLAKE_REDSHIFT,
    data=CASE_INSENSITIVE_DATA,
)
def test_quoted_success(batch_for_datasource: Batch) -> None:
    # When we create the table in the testing fr
    expectation = gxe.ExpectTableColumnsToMatchSet(
        column_set=['"column_a"', '"COLUMN_B"', '"CoLuMn_C"']
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=[SnowflakeDatasourceTestConfig()],
    data=CASE_INSENSITIVE_DATA,
)
def test_quoted_success_snowflake(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectTableColumnsToMatchSet(
        column_set=['"column_a"', '"CoLuMn_C"'],
        exact_match=False,
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=SQL_DATA_SOURCES,
    data=CASE_INSENSITIVE_DATA,
)
def test_quoted_failure(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectTableColumnsToMatchSet(
        column_set=['"Column_a"', '"COLUMN_B"', '"CoLuMn_C"']
    )
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@parameterize_batch_for_data_sources(
    data_source_configs=SQL_DATA_SOURCES_WITHOUT_SNOWFLAKE_REDSHIFT,
    data=CASE_INSENSITIVE_DATA,
)
def test_unquoted_and_quoted_success(batch_for_datasource: Batch) -> None:
    # Column A and B are case insensitve
    expectation = gxe.ExpectTableColumnsToMatchSet(
        column_set=["Column_a", "COLUMN_B", '"CoLuMn_C"']
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=SQL_DATA_SOURCES,
    data=CASE_INSENSITIVE_DATA,
)
def test_unquoted_and_quoted_with_unquoted_failure(batch_for_datasource: Batch) -> None:
    # Column A and B are case insensitve
    expectation = gxe.ExpectTableColumnsToMatchSet(
        column_set=["Column_az", "COLUMN_B", '"CoLuMn_C"']
    )
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@parameterize_batch_for_data_sources(
    data_source_configs=SQL_DATA_SOURCES,
    data=CASE_INSENSITIVE_DATA,
)
def test_unquoted_and_quoted_with_quoted_failure(batch_for_datasource: Batch) -> None:
    # Column C is case sensitve
    expectation = gxe.ExpectTableColumnsToMatchSet(
        column_set=["column_a", "COLUMN_B", '"Column_c"']
    )
    result = batch_for_datasource.validate(expectation)
    assert not result.success



from __future__ import annotations

import json
from typing import ClassVar, Type

from great_expectations.data_context.store.database_store_backend import (
    DatabaseStoreBackend,
)
from great_expectations.data_context.store.store import Store
from great_expectations.data_context.types.resource_identifiers import (
    ValidationMetricIdentifier,
)
from great_expectations.util import (
    load_class,
    verify_dynamic_loading_support,
)


class MetricStore(Store):
    """
    A MetricStore stores ValidationMetric information to be used between runs.
    """

    _key_class: ClassVar[Type] = ValidationMetricIdentifier

    def __init__(self, store_backend=None, store_name=None) -> None:
        if store_backend is not None:
            store_backend_module_name = store_backend.get(
                "module_name", "great_expectations.data_context.store"
            )
            store_backend_class_name = store_backend.get("class_name", "InMemoryStoreBackend")
            verify_dynamic_loading_support(module_name=store_backend_module_name)
            store_backend_class = load_class(store_backend_class_name, store_backend_module_name)

            if issubclass(store_backend_class, DatabaseStoreBackend):
                # Provide defaults for this common case
                if "table_name" not in store_backend:
                    store_backend["table_name"] = store_backend.get("table_name", "ge_metrics")
                if "key_columns" not in store_backend:
                    store_backend["key_columns"] = store_backend.get(
                        "key_columns",
                        [
                            "run_name",
                            "run_time",
                            "data_asset_name",
                            "expectation_suite_identifier",
                            "metric_name",
                            "metric_kwargs_id",
                        ],
                    )

        super().__init__(store_backend=store_backend, store_name=store_name)

    def serialize(self, value):  # type: ignore[explicit-override] # FIXME
        return json.dumps({"value": value})

    def deserialize(self, value):  # type: ignore[explicit-override] # FIXME
        if value:
            return json.loads(value)["value"]



import logging
import os
import shutil

import pytest

import great_expectations as gx
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.util import file_relative_path

pytestmark = pytest.mark.filesystem

logger = logging.getLogger(__name__)


def read_config_file_from_disk(config_filepath):
    with open(config_filepath) as infile:
        config_file = infile.read()
        return config_file


@pytest.fixture
def data_context_parameterized_expectation_suite_with_usage_statistics_enabled(
    tmp_path_factory,
):
    """
    This data_context is *manually* created to have the config we want, vs
    created with DataContext.create()
    """
    project_path = str(tmp_path_factory.mktemp("data_context"))
    context_path = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP
    asset_config_path = os.path.join(context_path, "expectations")  # noqa: PTH118 # FIXME CoP
    fixture_dir = file_relative_path(__file__, "../test_fixtures")
    os.makedirs(  # noqa: PTH103 # FIXME CoP
        os.path.join(asset_config_path, "my_dag_node"),  # noqa: PTH118 # FIXME CoP
        exist_ok=True,
    )
    shutil.copy(
        os.path.join(  # noqa: PTH118 # FIXME CoP
            fixture_dir, "great_expectations_v013_basic_with_usage_stats_enabled.yml"
        ),
        str(os.path.join(context_path, FileDataContext.GX_YML)),  # noqa: PTH118 # FIXME CoP
    )
    shutil.copy(
        os.path.join(  # noqa: PTH118 # FIXME CoP
            fixture_dir,
            "expectation_suites/parameterized_expectation_suite_fixture.json",
        ),
        os.path.join(asset_config_path, "my_dag_node", "default.json"),  # noqa: PTH118 # FIXME CoP
    )
    os.makedirs(  # noqa: PTH103 # FIXME CoP
        os.path.join(context_path, "plugins"),  # noqa: PTH118 # FIXME CoP
        exist_ok=True,
    )
    return gx.get_context(context_root_dir=context_path)


def test_preserve_comments_in_yml_after_adding_datasource(
    data_context_parameterized_expectation_suite_with_usage_statistics_enabled,
):
    # Skipping this test for now, because the order of the contents of the returned CommentedMap is inconsistent.  # noqa: E501 # FIXME CoP
    pytest.skip("KNOWN ISSUE")
    config_filepath = os.path.join(  # noqa: PTH118 # FIXME CoP
        data_context_parameterized_expectation_suite_with_usage_statistics_enabled.root_directory,
        FileDataContext.GX_YML,
    )
    initial_config = read_config_file_from_disk(config_filepath)
    print("++++++++++++++++ initial config +++++++++++++++++++++++")
    print(initial_config)
    print("----------------------------------------")

    data_context_parameterized_expectation_suite_with_usage_statistics_enabled.add_datasource(
        "test_datasource",
        module_name="great_expectations.datasource",
        class_name="PandasDatasource",
        batch_kwargs_generators={
            "subdir_reader": {
                "class_name": "SubdirReaderBatchKwargsGenerator",
                "base_directory": "../data",
            }
        },
    )

    # TODO The comments on lines 1,2 & 4 of the fixture exposes the bug.
    expected = """# This is a basic configuration for testing.
# It has comments that should be preserved.
config_version: 3.0
# Here's a comment between the config version and the datassources
datasources:
  # For example, this one.
  mydatasource: # This should stay by mydatasource
    module_name: great_expectations.datasource
    data_asset_type:
      class_name: PandasDataset
    batch_kwargs_generators:
      # The name default is read if no datasource or generator is specified
      mygenerator:
        class_name: SubdirReaderBatchKwargsGenerator
        base_directory: ../data
        reader_options:
          sep:
          engine: python
    class_name: PandasDatasource
  test_datasource:
    module_name: great_expectations.datasource
    batch_kwargs_generators:
      subdir_reader:
        class_name: SubdirReaderBatchKwargsGenerator
        base_directory: ../data
    data_asset_type:
      module_name: great_expectations.dataset
      class_name: PandasDataset
    class_name: PandasDatasource

config_variables_file_path: uncommitted/config_variables.yml

plugins_directory: plugins/
expectations_store_name: expectations_store
validation_results_store_name: validation_results_store
checkpoint_store_name: checkpoint_store

data_docs_sites:
  local_site:
    show_how_to_buttons: true
    class_name: SiteBuilder
    store_backend:
      class_name: TupleFilesystemStoreBackend
      base_directory: uncommitted/data_docs/local_site/
    site_index_builder:
      class_name: DefaultSiteIndexBuilder

stores:
  expectations_store:
    class_name: ExpectationsStore
    store_backend:
      class_name: TupleFilesystemStoreBackend
      base_directory: expectations/

  validation_results_store:
    class_name: ValidationResultsStore

  checkpoint_store:
    class_name: CheckpointStore
    store_backend:
      class_name: TupleFilesystemStoreBackend
      base_directory: checkpoints/

analytics_enabled: false
data_context_id: 7f76b3c9-330c-4307-b882-7ad9186adf0c



"""

    print("++++++++++++++++ expected +++++++++++++++++++++++")
    print(expected)
    print("----------------------------------------")

    observed = read_config_file_from_disk(config_filepath)

    print("++++++++++++++++ observed +++++++++++++++++++++++")
    print(observed)
    print("----------------------------------------")

    # TODO: this test fails now, but only on whitespace
    # Whitespace issues seem to come from the addition or deletion of keys
    assert observed.replace("\n", "") == expected.replace("\n", "")

    # What we really want:
    # assert observed == expected



from datetime import datetime, timezone

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.compatibility.bigquery import BIGQUERY_TYPES
from great_expectations.compatibility.postgresql import POSTGRESQL_TYPES
from great_expectations.compatibility.snowflake import SNOWFLAKE_TYPES
from great_expectations.compatibility.sqlalchemy import sqltypes
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import (
    BigQueryDatasourceTestConfig,
    DatabricksDatasourceTestConfig,
    MSSQLDatasourceTestConfig,
    MySQLDatasourceTestConfig,
    PandasDataFrameDatasourceTestConfig,
    PandasFilesystemCsvDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

try:
    from great_expectations.compatibility.pyspark import types as PYSPARK_TYPES

    spark_filesystem_csv_datasource_test_config = SparkFilesystemCsvDatasourceTestConfig(
        column_types={
            "created_at": PYSPARK_TYPES.TimestampType,
            "updated_at": PYSPARK_TYPES.DateType,
            "amount": PYSPARK_TYPES.FloatType,
            "quantity": PYSPARK_TYPES.IntegerType,
            "name": PYSPARK_TYPES.StringType,
        },
    )
except ModuleNotFoundError:
    spark_filesystem_csv_datasource_test_config = SparkFilesystemCsvDatasourceTestConfig()

DATA = pd.DataFrame(
    {
        "created_at": [
            datetime(year=2021, month=1, day=30, tzinfo=timezone.utc),
            datetime(year=2022, month=1, day=30, tzinfo=timezone.utc),
            datetime(year=2023, month=1, day=30, tzinfo=timezone.utc),
        ],
        "updated_at": [
            datetime(year=2021, month=1, day=31, tzinfo=timezone.utc).date(),
            datetime(year=2022, month=1, day=31, tzinfo=timezone.utc).date(),
            datetime(year=2023, month=1, day=31, tzinfo=timezone.utc).date(),
        ],
        "amount": [1.00, 2.00, 3.00],
        "quantity": [1, 2, 3],
        "name": ["albert", "issac", "galileo"],
    }
)


# some backends fail to load datetimes into the database unless they are strings
DATA_WITH_STRING_DATETIMES = pd.DataFrame(
    {
        "created_at": [
            str(datetime(year=2021, month=1, day=30, tzinfo=timezone.utc)),
            str(datetime(year=2022, month=1, day=30, tzinfo=timezone.utc)),
            str(datetime(year=2023, month=1, day=30, tzinfo=timezone.utc)),
        ],
        "updated_at": [
            datetime(year=2021, month=1, day=31, tzinfo=timezone.utc).date(),
            datetime(year=2022, month=1, day=31, tzinfo=timezone.utc).date(),
            datetime(year=2023, month=1, day=31, tzinfo=timezone.utc).date(),
        ],
        "amount": [1.00, 2.00, 3.00],
        "quantity": [1, 2, 3],
        "name": ["albert", "issac", "galileo"],
    }
)


PANDAS_TEST_CASES = [
    pytest.param(
        'name=="albert"',
        id="text-eq",
    ),
    pytest.param(
        "quantity<3",
        id="number-lt",
    ),
    pytest.param(
        "quantity==1",
        id="number-eq",
    ),
    pytest.param(
        "updated_at<datetime.date(2021,2,1)",
        id="datetime.date-lt",
    ),
    pytest.param(
        "updated_at>datetime.date(2021,1,30)",
        id="datetime.date-gt",
    ),
    pytest.param(
        "updated_at==datetime.date(2021,1,31)",
        id="datetime.date-eq",
    ),
    pytest.param(
        "created_at<datetime.datetime(2021,1,31,0,0,0,tzinfo=datetime.timezone.utc)",
        id="datetime.datetime-lt",
    ),
    pytest.param(
        "created_at>datetime.datetime(2021,1,29,0,0,0,tzinfo=datetime.timezone.utc)",
        id="datetime.datetime-gt",
    ),
    pytest.param(
        "created_at==datetime.datetime(2021,1,30,0,0,0,tzinfo=datetime.timezone.utc)",
        id="datetime.datetime-eq",
    ),
]


@parameterize_batch_for_data_sources(
    data_source_configs=[
        PandasDataFrameDatasourceTestConfig(),
        PandasFilesystemCsvDatasourceTestConfig(
            read_options={
                "parse_dates": ["created_at", "updated_at"],
                "date_format": "mixed",
            },
        ),
    ],
    data=DATA,
)
@pytest.mark.parametrize(
    "row_condition",
    PANDAS_TEST_CASES,
)
def test_expect_column_min_to_be_between__pandas_row_condition(
    batch_for_datasource: Batch, row_condition: str
) -> None:
    expectation = gxe.ExpectColumnMinToBeBetween(
        column="amount",
        min_value=0.5,
        max_value=1.5,
        row_condition=row_condition,
        condition_parser="pandas",
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


SQL_AND_SPARK_TEST_CASES = [
    pytest.param(
        'col("name")=="albert"',
        id="text-eq",
    ),
    pytest.param(
        'col("quantity")<3',
        id="number-lt",
    ),
    pytest.param(
        'col("quantity")==1',
        id="number-eq",
    ),
    pytest.param(
        'col("updated_at")<date("2021-02-01"))',
        id="date-lt",
    ),
    pytest.param(
        'col("updated_at")>date("2021-01-30"))',
        id="date-gt",
    ),
    pytest.param(
        'col("updated_at")==date("2021-01-31"))',
        id="date-eq",
    ),
    pytest.param(
        'col("created_at")<date("2021-01-31 00:00:00"))',
        id="datetime-lt",
    ),
    pytest.param(
        'col("created_at")>date("2021-01-29 00:00:00"))',
        id="datetime-gt",
    ),
]


@parameterize_batch_for_data_sources(
    data_source_configs=[
        BigQueryDatasourceTestConfig(
            column_types={
                "created_at": BIGQUERY_TYPES.DATETIME,
                "updated_at": BIGQUERY_TYPES.DATE,
            }
        ),
        MSSQLDatasourceTestConfig(),
        MySQLDatasourceTestConfig(
            column_types={
                "created_at": sqltypes.TIMESTAMP(timezone=True),
                "updated_at": sqltypes.DATE,
            }
        ),
        PostgreSQLDatasourceTestConfig(
            column_types={
                "created_at": POSTGRESQL_TYPES.TIMESTAMP,
                "updated_at": POSTGRESQL_TYPES.DATE,
            }
        ),
        SqliteDatasourceTestConfig(),
    ],
    data=DATA,
)
@pytest.mark.parametrize(
    "row_condition",
    SQL_AND_SPARK_TEST_CASES,
)
def test_expect_column_min_to_be_between__sql_row_condition(
    batch_for_datasource: Batch, row_condition: str
) -> None:
    expectation = gxe.ExpectColumnMinToBeBetween(
        column="amount",
        min_value=0.5,
        max_value=1.5,
        row_condition=row_condition,
        condition_parser="great_expectations",
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=[
        DatabricksDatasourceTestConfig(),
        SnowflakeDatasourceTestConfig(
            column_types={
                "created_at": SNOWFLAKE_TYPES.TIMESTAMP_TZ,
                "updated_at": sqltypes.DATE,  # snowflake.sqlalchemy missing snowflake DATE type
            }
        ),
    ],
    data=DATA_WITH_STRING_DATETIMES,
)
@pytest.mark.parametrize(
    "row_condition",
    SQL_AND_SPARK_TEST_CASES,
)
def test_expect_column_min_to_be_between__snowflake_databricks_row_condition(
    batch_for_datasource: Batch, row_condition: str
) -> None:
    expectation = gxe.ExpectColumnMinToBeBetween(
        column="amount",
        min_value=0.5,
        max_value=1.5,
        row_condition=row_condition,
        condition_parser="great_expectations",
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=[spark_filesystem_csv_datasource_test_config],
    data=DATA,
)
@pytest.mark.parametrize(
    "row_condition",
    SQL_AND_SPARK_TEST_CASES,
)
def test_expect_column_min_to_be_between__spark_row_condition(
    batch_for_datasource: Batch, row_condition: str
) -> None:
    expectation = gxe.ExpectColumnMinToBeBetween(
        column="amount",
        min_value=0.5,
        max_value=1.5,
        row_condition=row_condition,
        condition_parser="great_expectations",
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, TypeVar

from typing_extensions import ParamSpec

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.expectation_validation_result import (
    ExpectationValidationResult,
)
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)

if TYPE_CHECKING:
    from great_expectations.checkpoint.checkpoint import CheckpointResult

P = ParamSpec("P")
T = TypeVar("T")


def renderer(renderer_type: str, **kwargs) -> Callable[[Callable[P, T]], Callable[P, T]]:
    def wrapper(renderer_fn: Callable[P, T]) -> Callable[P, T]:
        @wraps(renderer_fn)
        def inner_func(*args: P.args, **kwargs: P.kwargs):
            return renderer_fn(*args, **kwargs)

        inner_func._renderer_type = renderer_type  # type: ignore[attr-defined] # FIXME CoP
        inner_func._renderer_definition_kwargs = kwargs  # type: ignore[attr-defined] # FIXME CoP
        return inner_func

    return wrapper


@public_api
class Renderer:
    """A convenience class to provide an explicit mechanism to instantiate any Renderer."""

    @override
    def __eq__(self, other: object) -> bool:
        # Renderers do not have any state, so they are equal if they are the same class
        return type(self) is type(other)

    @override
    def __hash__(self) -> int:
        return hash(type(self))

    def serialize(self) -> dict:
        # Necessary to enable proper serialization within an Action (and additionally, within a Checkpoint)  # noqa: E501 # FIXME CoP
        # TODO: Renderers should be ported over to Pydantic to prevent this fork in logic
        return {
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }

    @classmethod
    def _get_expectation_type(cls, ge_object):
        if isinstance(ge_object, ExpectationConfiguration):
            return ge_object.type

        elif isinstance(ge_object, ExpectationValidationResult):
            # This is a validation
            return ge_object.expectation_config.type

    # TODO: When we implement a ValidationResultSuite class, this method will move there.
    @classmethod
    def _find_evr_by_type(cls, evrs, type_):
        for evr in evrs:
            if evr.expectation_config.type == type_:
                return evr

    # TODO: When we implement a ValidationResultSuite class, this method will move there.
    @classmethod
    def _find_all_evrs_by_type(cls, evrs, type_, column_=None):
        ret = []
        for evr in evrs:
            if evr.expectation_config.type == type_ and (
                not column_ or column_ == evr.expectation_config.kwargs.get("column")
            ):
                ret.append(evr)

        return ret

    # TODO: When we implement a ValidationResultSuite class, this method will move there.
    @classmethod
    def _get_column_list_from_evrs(cls, evrs):
        """
        Get list of column names.

        If expect_table_columns_to_match_ordered_list EVR is present, use it as the list, including the order.

        Otherwise, get the list of all columns mentioned in the expectations and order it alphabetically.

        :param evrs:
        :return: list of columns with best effort sorting
        """  # noqa: E501 # FIXME CoP
        evrs_ = evrs if isinstance(evrs, list) else evrs.results

        expect_table_columns_to_match_ordered_list_evr = cls._find_evr_by_type(
            evrs_, "expect_table_columns_to_match_ordered_list"
        )
        # Group EVRs by column
        sorted_columns = sorted(
            list(
                {
                    evr.expectation_config.kwargs["column"]
                    for evr in evrs_
                    if "column" in evr.expectation_config.kwargs
                }
            )
        )

        if expect_table_columns_to_match_ordered_list_evr:
            ordered_columns = expect_table_columns_to_match_ordered_list_evr.result[
                "observed_value"
            ]
        else:
            ordered_columns = []

        # only return ordered columns from expect_table_columns_to_match_ordered_list evr if they match set of column  # noqa: E501 # FIXME CoP
        # names from entire evr
        if set(sorted_columns) == set(ordered_columns):
            return ordered_columns
        else:
            return sorted_columns

    # TODO: When we implement a ValidationResultSuite class, this method will move there.
    @classmethod
    def _group_evrs_by_column(cls, validation_results):
        columns = {}
        for evr in validation_results.results:
            if "column" in evr.expectation_config.kwargs:
                column = evr.expectation_config.kwargs["column"]
            else:
                column = "Table-level Expectations"

            if column not in columns:
                columns[column] = []
            columns[column].append(evr)

        return columns

    def render(self, checkpoint_result: CheckpointResult) -> Any:
        """
        Render interface method.
        """
        raise NotImplementedError



import pandas as pd
import pytest

from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.column.descriptive_stats import (
    ColumnDescriptiveStats,
    ColumnDescriptiveStatsResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import ALL_DATA_SOURCES

COLUMN_NAME = "whatevs"
DATA_FRAME = pd.DataFrame(
    {
        COLUMN_NAME: [1, 1, 2, 2, 3, 3, 4, 4, 5, 5],
    },
)


class TestColumnDescriptiveStats:
    @parameterize_batch_for_data_sources(
        data_source_configs=ALL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_descriptive_stats(self, batch_for_datasource: Batch) -> None:
        metric = ColumnDescriptiveStats(column=COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnDescriptiveStatsResult)
        assert metric_result.value.min == 1
        assert metric_result.value.max == 5
        assert metric_result.value.mean == pytest.approx(3)
        assert metric_result.value.standard_deviation == pytest.approx(1.4907119849998598)



from __future__ import annotations

from typing import Any, ClassVar, Dict, List, Literal, Tuple, Type, Union

from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.expectations.expectation import (
    ColumnPairMapExpectation,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import (
    COLUMN_A_DESCRIPTION,
    COLUMN_B_DESCRIPTION,
    MOSTLY_DESCRIPTION,
)

EXPECTATION_SHORT_DESCRIPTION = (
    "Expect the paired values from columns A and B to belong to a set of valid pairs."
)
VALUE_PAIRS_SET_DESCRIPTION = "All the valid pairs to be matched."
DATA_QUALITY_ISSUES = [DataQualityIssues.VALIDITY.value]

SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.REDSHIFT.value,
]


class ExpectColumnPairValuesToBeInSet(ColumnPairMapExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectColumnPairValuesToBeInSet is a \
    Column Pair Map Expectation.

    Column Pair Map Expectations are evaluated for a pair of columns and ask a yes/no question about the row-wise relationship between those two columns.
    Based on the result, they then calculate the percentage of rows that gave a positive answer.
    If the percentage is high enough, the Expectation considers that data valid.

    Args:
        column_A (str): {COLUMN_A_DESCRIPTION}
        column_B (str): {COLUMN_B_DESCRIPTION}
        value_pairs_set (list of tuples): {VALUE_PAIRS_SET_DESCRIPTION}

    Other Parameters:
        ignore_row_if (str): \
            "both_values_are_missing", "either_value_is_missing", "neither" \
            If specified, sets the condition on which a given row is to be ignored. Default "neither".
        mostly (None or a float between 0 and 1): \
            {MOSTLY_DESCRIPTION} \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly). Default 1.
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without modification. \
            For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	1       1
            1 	2       1
            2 	4   	1

    Code Examples:
        Passing Case:
            Input:
                ExpectColumnPairValuesToBeInSet(
                    column_A="test",
                    column_B="test2",
                    value_pairs_set=[(2,1), (1,1)],
                    mostly=.5
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 1,
                    "unexpected_percent": 33.33333333333333,
                    "partial_unexpected_list": [
                      [
                        4,
                        1
                      ]
                    ],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 33.33333333333333,
                    "unexpected_percent_nonmissing": 33.33333333333333
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectColumnPairValuesToBeInSet(
                    column_A="test",
                    column_B="test2",
                    value_pairs_set=[(1,2) (4,1)],
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 2,
                    "unexpected_percent": 66.66666666666666,
                    "partial_unexpected_list": [
                      [
                        1,
                        1
                      ],
                      [
                        2,
                        1
                      ]
                    ],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 66.66666666666666,
                    "unexpected_percent_nonmissing": 66.66666666666666
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    value_pairs_set: Union[List[Tuple[Any, Any]], SuiteParameterDict]
    ignore_row_if: Union[
        Literal["both_values_are_missing", "either_value_is_missing", "neither"],
        SuiteParameterDict,
    ] = "both_values_are_missing"

    # This dictionary contains metadata for display in the public gallery
    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": [
            "core expectation",
            "column pair map expectation",
        ],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    map_metric = "column_pair_values.in_set"
    success_keys: ClassVar[Tuple[str, ...]] = (
        "value_pairs_set",
        "ignore_row_if",
        "mostly",
    )
    args_keys = (
        "column_A",
        "column_B",
        "value_pairs_set",
    )

    class Config:
        title = "Expect column pair values to be in set"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectColumnPairValuesToBeInSet]
        ) -> None:
            ColumnPairMapExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )
